# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Byte Stream Conversion Tests."""

from __future__ import annotations

from rosbags.typesys import Stores, get_types_from_msg, get_typestore

STATIC_16_64 = """
uint16 u16
uint64 u64
"""

DYNAMIC_S_64 = """
string s
uint64 u64
"""


def test_ros1_to_cdr() -> None:
    """Test ROS1 to CDR conversion."""
    store = get_typestore(Stores.LATEST)

    msgtype = 'test_msgs/msg/static_16_64'
    store.register(dict(get_types_from_msg(STATIC_16_64, msgtype)))
    msg_ros = b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x02'
    msg_cdr = b'\x00\x01\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02'
    assert store.ros1_to_cdr(msg_ros, msgtype) == msg_cdr
    assert store.serialize_cdr(store.deserialize_ros1(msg_ros, msgtype), msgtype) == msg_cdr

    msgtype = 'test_msgs/msg/dynamic_s_64'
    store.register(dict(get_types_from_msg(DYNAMIC_S_64, msgtype)))
    msg_ros = b'\x01\x00\x00\x00X\x00\x00\x00\x00\x00\x00\x00\x02'
    msg_cdr = b'\x00\x01\x00\x00\x02\x00\x00\x00X\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02'
    assert store.ros1_to_cdr(msg_ros, msgtype) == msg_cdr
    assert store.serialize_cdr(store.deserialize_ros1(msg_ros, msgtype), msgtype) == msg_cdr


def test_cdr_to_ros1() -> None:
    """Test CDR to ROS1 conversion."""
    store = get_typestore(Stores.LATEST)

    msgtype = 'test_msgs/msg/static_16_64'
    store.register(dict(get_types_from_msg(STATIC_16_64, msgtype)))
    msg_ros = b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x02'
    msg_cdr = b'\x00\x01\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02'
    assert store.cdr_to_ros1(msg_cdr, msgtype) == msg_ros
    assert store.serialize_ros1(store.deserialize_cdr(msg_cdr, msgtype), msgtype) == msg_ros

    msgtype = 'test_msgs/msg/dynamic_s_64'
    store.register(dict(get_types_from_msg(DYNAMIC_S_64, msgtype)))
    msg_ros = b'\x01\x00\x00\x00X\x00\x00\x00\x00\x00\x00\x00\x02'
    msg_cdr = b'\x00\x01\x00\x00\x02\x00\x00\x00X\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02'
    assert store.cdr_to_ros1(msg_cdr, msgtype) == msg_ros
    assert store.serialize_ros1(store.deserialize_cdr(msg_cdr, msgtype), msgtype) == msg_ros

    Header = store.types['std_msgs/msg/Header']  # noqa: N806
    Time = store.types['builtin_interfaces/msg/Time']  # noqa: N806

    header = Header(stamp=Time(42, 666), frame_id='frame')
    msg_ros = store.cdr_to_ros1(
        store.serialize_cdr(header, 'std_msgs/msg/Header'), 'std_msgs/msg/Header'
    ).tobytes()
    assert msg_ros == b'\x00\x00\x00\x00*\x00\x00\x00\x9a\x02\x00\x00\x05\x00\x00\x00frame'



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Rosbags Type System.

The type system manages ROS message types and ships all standard ROS2
distribution message types by default. The system supports custom message
types through parsers that dynamically parse custom message definitions
from different source formats.

Supported formats:
  - IDL files (subset of the standard necessary for parsing ROS2 IDL) `[1]`_
  - MSG files `[2]`_

.. _[1]: https://www.omg.org/spec/IDL/About-IDL/
.. _[2]: http://wiki.ros.org/msg

"""

from .base import TypesysError
from .idl import get_types_from_idl
from .msg import get_types_from_msg
from .stores import Stores, get_typestore

__all__ = [
    'Stores',
    'TypesysError',
    'get_types_from_idl',
    'get_types_from_msg',
    'get_typestore',
]



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Rosbags typing."""

import sys
from enum import IntEnum, auto
from typing import Any, BinaryIO, Literal, Protocol, TypeAlias, TypeVar

if sys.version_info >= (3, 11):
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import Self

T = TypeVar('T')


class Nodetype(IntEnum):
    """Parse tree node types.

    The first four match the Valtypes of final message definitions.
    """

    BASE = auto()
    NAME = auto()
    ARRAY = auto()
    SEQUENCE = auto()


Basename: TypeAlias = Literal[
    'bool',
    'byte',
    'char',
    'int8',
    'int16',
    'int32',
    'int64',
    'uint8',
    'uint16',
    'uint32',
    'uint64',
    'float32',
    'float64',
    'float128',
    'string',
]
Basetype: TypeAlias = tuple[Basename, int]

BaseDesc: TypeAlias = tuple[Literal[Nodetype.BASE], Basetype]
NameDesc: TypeAlias = tuple[Literal[Nodetype.NAME], str]
FieldDesc: TypeAlias = (
    BaseDesc
    | NameDesc
    | tuple[Literal[Nodetype.ARRAY, Nodetype.SEQUENCE], tuple[BaseDesc | NameDesc, int]]
)

ConstValue: TypeAlias = str | bool | int | float

Constdefs: TypeAlias = list[tuple[str, Basename, ConstValue]]
Fielddefs: TypeAlias = list[tuple[str, FieldDesc]]
Typesdict: TypeAlias = dict[str, tuple[Constdefs, Fielddefs]]


class StatResult(Protocol):  # pragma: no cover
    """Start result protocol."""

    @property
    def st_size(self) -> int:
        """Proxy."""
        raise NotImplementedError


class RPath(Protocol):  # pragma: no cover
    """Reader path protocol."""

    def exists(self) -> bool:
        """Proxy."""
        raise NotImplementedError

    def is_dir(self) -> bool:
        """Proxy."""
        raise NotImplementedError

    def open(  # type: ignore[explicit-any]
        self,
        *args: Any,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> BinaryIO:
        """Proxy."""
        raise NotImplementedError

    def read_text(
        self,
        encoding: str | None = None,
    ) -> str:
        """Proxy."""
        raise NotImplementedError

    def stat(self, *, follow_symlinks: bool = True) -> StatResult:
        """Proxy."""
        raise NotImplementedError

    def __truediv__(self, key: str | Self) -> Self:
        """Proxy."""
        raise NotImplementedError

    @property
    def stem(self) -> str:
        """Proxy."""
        raise NotImplementedError

    @property
    def suffix(self) -> str:
        """Proxy."""
        raise NotImplementedError



"""Example: Save images as rosbag1."""

from __future__ import annotations

import numpy as np

from rosbags.rosbag1 import Writer
from rosbags.typesys import Stores, get_typestore
from rosbags.typesys.stores.ros1_noetic import (
    builtin_interfaces__msg__Time as Time,
    sensor_msgs__msg__CompressedImage as CompressedImage,
    std_msgs__msg__Header as Header,
)

TOPIC = '/camera'
FRAMEID = 'map'

# Contains filenames and their timestamps
IMAGES = [('homer.jpg', 42), ('marge.jpg', 43)]


def save_images() -> None:
    """Iterate over IMAGES and save to output bag."""
    typestore = get_typestore(Stores.ROS1_NOETIC)
    with Writer('output.bag') as writer:
        conn = writer.add_connection(TOPIC, CompressedImage.__msgtype__, typestore=typestore)

        for idx, (path, timestamp) in enumerate(IMAGES):
            msg = CompressedImage(
                Header(
                    seq=idx,
                    stamp=Time(sec=int(timestamp // 10**9), nanosec=int(timestamp % 10**9)),
                    frame_id=FRAMEID,
                ),
                format='jpeg',  # could also be 'png'
                data=np.fromfile(path, dtype=np.uint8),
            )

            writer.write(
                conn,
                timestamp,
                typestore.serialize_ros1(msg, CompressedImage.__msgtype__),
            )



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Test Fixturest."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, cast
from unittest.mock import patch

import numpy as np
import pytest

if sys.version_info >= (3, 12):  # pragma: no cover
    from typing import override
else:  # pragma: no cover
    from typing_extensions import override

if TYPE_CHECKING:
    from collections.abc import Generator


@pytest.fixture
def _comparable() -> Generator[None, None, None]:
    """Make messages containing numpy arrays comparable.

    Notes:
        This solution is necessary as numpy.ndarray is not directly patchable.

    """
    frombuffer = np.frombuffer

    class CNDArray:
        """Comparable ndarray."""

        def __init__(self, child: np.ndarray[tuple[int, ...], np.dtype[np.uint8]]) -> None:
            self.child = child

        @override
        def __eq__(self, other: object) -> bool:
            if isinstance(other, CNDArray):
                other = other.child
            assert isinstance(other, np.ndarray)
            return bool(
                np.equal(
                    self.child,
                    cast('np.ndarray[tuple[int, ...], np.dtype[np.uint8]]', other),
                ).all(),
            )

        def byteswap(self) -> CNDArray:
            return CNDArray(self.child.byteswap())

        def __deepcopy__(self, _: dict[str, object]) -> CNDArray:
            return self

        def __getattr__(self, name: str) -> object:
            return cast('object', getattr(self.child, name))

        def __hash__(self) -> int:
            return hash(self.child)

    def wrap_frombuffer(mem: bytes, dtype: str, count: int, offset: int) -> CNDArray:
        return CNDArray(frombuffer(mem, dtype=dtype, count=count, offset=offset))

    with patch.object(np, 'frombuffer', wrap_frombuffer):
        yield



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Mcap storage."""

from __future__ import annotations

import heapq
import struct
from collections import defaultdict
from dataclasses import dataclass
from importlib.metadata import version
from io import BytesIO
from struct import iter_unpack, unpack_from
from typing import TYPE_CHECKING, NamedTuple, cast

import zstandard
from lz4.frame import decompress as lz4_decompress  # type: ignore[import-untyped]

from rosbags.interfaces import (
    Connection,
    ConnectionExtRosbag2,
    MessageDefinition,
    MessageDefinitionFormat,
    Qos,
)

from .enums import CompressionMode
from .errors import ReaderError
from .metadata import ReaderMetadata, parse_qos

if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Iterable
    from pathlib import Path
    from typing import BinaryIO

    from rosbags.interfaces.typing import RPath

    Unpack = Callable[[bytes], 'tuple[int]']
    Unpack2 = Callable[[bytes], 'tuple[int, int]']
    Unpack4 = Callable[[bytes], 'tuple[int, int, int, int]']
    Unpack5 = Callable[[bytes], 'tuple[int, int, int, int, int]']


class Schema(NamedTuple):
    """Schema."""

    id: int
    name: str
    encoding: str
    data: str


class Channel(NamedTuple):
    """Channel."""

    id: int
    schema: str
    topic: str
    message_encoding: str
    metadata: bytes  # dict[str, str]


class Chunk(NamedTuple):
    """Chunk."""

    start_time: int
    end_time: int
    size: int
    crc: int
    compression: str
    records: bytes


class ChunkInfo(NamedTuple):
    """Chunk."""

    message_start_time: int
    message_end_time: int
    chunk_start_offset: int
    chunk_length: int
    message_index_offsets: dict[int, int]
    message_index_length: int
    compression: str
    compressed_size: int
    uncompressed_size: int
    channel_count: dict[int, int]


class Statistics(NamedTuple):
    """Statistics."""

    message_count: int
    schema_count: int
    channel_count: int
    attachement_count: int
    metadata_count: int
    chunk_count: int
    start_time: int
    end_time: int
    channel_message_counts: dict[int, int]


class Msg(NamedTuple):
    """Message wrapper."""

    timestamp: int
    offset: int
    connection: Connection | None
    data: bytes | None


MAXSIZE: int = 2**63 - 1

deserialize_uint16: Unpack = struct.Struct('<H').unpack
deserialize_uint32: Unpack = struct.Struct('<I').unpack
deserialize_uint64: Unpack = struct.Struct('<Q').unpack

deserialize_hq: Unpack2 = struct.Struct('<HQ').unpack
deserialize_qq: Unpack2 = struct.Struct('<QQ').unpack
deserialize_qqqi: Unpack4 = struct.Struct('<QQQI').unpack
deserialize_qqqq: Unpack4 = struct.Struct('<QQQQ').unpack
deserialize_hiqq: Unpack4 = struct.Struct('<HIQQ').unpack
deserialize_qhiqq: Unpack5 = struct.Struct('<QHIQQ').unpack


def read_sized(bio: BinaryIO) -> bytes:
    """Read one record."""
    return bio.read(deserialize_uint64(bio.read(8))[0])


def skip_sized(bio: BinaryIO) -> None:
    """Read one record."""
    _ = bio.seek(deserialize_uint64(bio.read(8))[0], 1)


def read_bytes(bio: BinaryIO) -> bytes:
    """Read string."""
    return bio.read(deserialize_uint32(bio.read(4))[0])


def read_string(bio: BinaryIO) -> str:
    """Read string."""
    return bio.read(deserialize_uint32(bio.read(4))[0]).decode()


DECOMPRESSORS: dict[str, Callable[[bytes, int], bytes]] = {
    '': lambda x, _: x,
    'lz4': lambda x, _: lz4_decompress(x),
    'zstd': zstandard.ZstdDecompressor().decompress,
}


def msgsrc(
    chunk: ChunkInfo,
    channel_map: dict[int, Connection],
    start: int,
    stop: int,
    bio: BinaryIO,
) -> Generator[Msg, None, None]:
    """Yield messages from chunk in time order."""
    yield Msg(chunk.message_start_time, 0, None, None)

    _ = bio.seek(chunk.chunk_start_offset + 9 + 40 + len(chunk.compression))
    compressed_data = bio.read(chunk.compressed_size)
    subio = BytesIO(DECOMPRESSORS[chunk.compression](compressed_data, chunk.uncompressed_size))

    messages: list[Msg] = []
    while (offset := subio.tell()) < chunk.uncompressed_size:
        op_ = ord(subio.read(1))
        if op_ == 0x05:
            recio = BytesIO(read_sized(subio))
            channel_id, _, log_time, _ = deserialize_hiqq(recio.read(22))
            if start <= log_time < stop and channel_id in channel_map:
                messages.append(
                    Msg(
                        log_time,
                        chunk.chunk_start_offset + offset,
                        channel_map[channel_id],
                        recio.read(),
                    ),
                )
        else:
            skip_sized(subio)

    yield from sorted(messages, key=lambda x: x.timestamp)


class McapReader:
    """Mcap format reader."""

    def __init__(self, path: RPath) -> None:
        """Initialize."""
        self.path = path
        self.bio: BinaryIO | None = None
        self.data_start = 0
        self.data_end = 0
        self.schemas: dict[int, Schema] = {}
        self.channels: dict[int, Channel] = {}
        self.chunks: list[ChunkInfo] = []
        self.statistics: Statistics | None = None
        self.connections: list[Connection] = []
        self.metadata = ReaderMetadata(0, 2**63 - 1, 0, 0, None, None, None, None)

    def open(self) -> None:
        """Open MCAP."""
        try:
            self.bio = self.path.open('rb')
        except OSError as err:
            msg = f'Could not open file {str(self.path)!r}: {err.strerror}.'
            raise ReaderError(msg) from err

        magic = self.bio.read(8)
        if not magic:
            msg = f'File {str(self.path)!r} seems to be empty.'
            raise ReaderError(msg)

        if magic != b'\x89MCAP0\r\n':
            msg = 'File magic is invalid.'
            raise ReaderError(msg)

        op_ = ord(self.bio.read(1))
        if op_ != 0x01:
            msg = 'Unexpected record.'
            raise ReaderError(msg)

        recio = BytesIO(read_sized(self.bio))
        profile = read_string(recio)
        if profile != 'ros2':
            msg = 'Profile is not ros2.'
            raise ReaderError(msg)
        self.data_start = self.bio.tell()

        _ = self.bio.seek(-37, 2)
        footer_start = self.bio.tell()
        data = self.bio.read()
        magic = data[-8:]
        if magic != b'\x89MCAP0\r\n':
            msg = 'File end magic is invalid.'
            raise ReaderError(msg)

        assert len(data) == 37
        assert data[0:9] == b'\x02\x14\x00\x00\x00\x00\x00\x00\x00', data[0:9]

        (summary_start,) = deserialize_uint64(data[9:17])
        if summary_start:
            self.data_end = summary_start
            self.read_index()
            if self.statistics:
                if not self.schemas:
                    self.meta_scan()
            elif self.chunks:
                message_count = sum(sum(x.channel_count.values()) for x in self.chunks)
                start_time = min(x.message_start_time for x in self.chunks)
                end_time = max(x.message_end_time for x in self.chunks)
                duration = end_time - start_time
                cstats: dict[int, int] = defaultdict(int)
                for chunk in self.chunks:
                    for cid, count in chunk.channel_count.items():
                        cstats[cid] += count
                self.statistics = Statistics(
                    message_count,
                    len(self.schemas),
                    len(self.channels),
                    0,
                    0,
                    len(self.chunks),
                    start_time,
                    end_time,
                    cstats,
                )
            else:
                self.meta_scan()
        else:
            self.data_end = footer_start
            self.meta_scan()

        def get_msgdef(name: str) -> MessageDefinition:
            """Get message definition for name."""
            fmtmap = {
                'ros2msg': MessageDefinitionFormat.MSG,
                'ros2idl': MessageDefinitionFormat.IDL,
                'omgidl': MessageDefinitionFormat.IDL,
                '': MessageDefinitionFormat.NONE,
            }
            if msgtype := next((x for x in self.schemas.values() if x.name == name), None):
                return MessageDefinition(fmtmap[msgtype.encoding], msgtype.data)
            return MessageDefinition(MessageDefinitionFormat.NONE, '')

        def get_qos(metadata: bytes) -> list[Qos]:
            bio = BytesIO(metadata)
            while bio.tell() < len(metadata):
                key = read_string(bio)
                value = read_string(bio)
                if key == 'offered_qos_profiles':
                    return parse_qos(value)
            return []

        assert self.statistics
        self.connections = [
            Connection(
                x.id,
                x.topic,
                x.schema,
                get_msgdef(x.schema),
                '',
                self.statistics.channel_message_counts.get(x.id, 0),
                ConnectionExtRosbag2(
                    x.message_encoding,
                    get_qos(x.metadata),
                ),
                self,
            )
            for x in self.channels.values()
        ]

        message_count = self.statistics.message_count
        start_time = self.statistics.start_time
        end_time = self.statistics.end_time
        duration = end_time - start_time

        self.metadata = self.metadata._replace(
            duration=duration + 1,
            start_time=start_time,
            end_time=end_time + 1,
            message_count=message_count,
        )

    def read_index(self) -> None:
        """Read index from file."""
        bio = self.bio
        assert bio

        schemas = self.schemas
        channels = self.channels
        chunks = self.chunks

        _ = bio.seek(self.data_end)
        while True:
            op_ = ord(bio.read(1))

            if op_ in {0x02, 0x0E}:
                break

            if op_ == 0x03:
                _ = bio.seek(8, 1)
                (key,) = deserialize_uint16(bio.read(2))
                schemas[key] = Schema(key, read_string(bio), read_string(bio), read_string(bio))

            elif op_ == 0x04:
                _ = bio.seek(8, 1)
                (key,) = deserialize_uint16(bio.read(2))
                schema_name = schemas.get(
                    deserialize_uint16(bio.read(2))[0],
                    Schema(0, '__schemaless__', 'cdr', ''),
                ).name
                channels[key] = Channel(
                    key,
                    schema_name,
                    read_string(bio),
                    read_string(bio),
                    read_bytes(bio),
                )

            elif op_ == 0x08:
                _ = bio.seek(8, 1)
                chunk = ChunkInfo(
                    *deserialize_qqqq(bio.read(32)),
                    {
                        x[0]: x[1]
                        for x in cast(
                            'Iterable[tuple[int, int]]',
                            iter_unpack('<HQ', bio.read(deserialize_uint32(bio.read(4))[0])),
                        )
                    },
                    *deserialize_uint64(bio.read(8)),
                    read_string(bio),
                    *deserialize_qq(bio.read(16)),
                    {},
                )
                offset_channel = sorted((v, k) for k, v in chunk.message_index_offsets.items())
                offsets = [
                    *[x[0] for x in offset_channel],
                    chunk.chunk_start_offset + chunk.chunk_length + chunk.message_index_length,
                ]
                chunk.channel_count.update(
                    {
                        x[1]: count // 16
                        for x, y, z in zip(offset_channel, offsets[1:], offsets, strict=False)
                        if (count := y - z - 15)
                    },
                )
                chunks.append(chunk)

            elif op_ == 0x0A:
                skip_sized(bio)

            elif op_ == 0x0B:
                _ = bio.seek(8, 1)
                self.statistics = Statistics(
                    *cast(
                        'tuple[int, int, int, int, int, int, int ,int]',
                        unpack_from('<QHIIIIQQ', bio.read(42), 0),
                    ),
                    dict(
                        deserialize_hq(bio.read(10))
                        for _ in range(deserialize_uint32(bio.read(4))[0] // 10)
                    ),
                )

            elif op_ == 0x0D:
                skip_sized(bio)

            else:
                skip_sized(bio)

    def close(self) -> None:
        """Close MCAP."""
        assert self.bio
        self.bio.close()
        self.bio = None

    def meta_scan(self) -> None:
        """Generate metadata by scanning through file."""
        assert self.bio
        bio = self.bio
        bio_size = self.data_end
        _ = bio.seek(self.data_start)

        msgcount = 0
        start_time = 2**63 - 1
        end_time = 0
        nchunks = 0
        cstats: dict[int, int] = defaultdict(int)

        schemas = self.schemas
        channels = self.channels

        while bio.tell() < bio_size:
            op_ = ord(bio.read(1))

            if op_ == 0x03:
                _ = bio.seek(8, 1)
                (key,) = deserialize_uint16(bio.read(2))
                schemas[key] = Schema(key, read_string(bio), read_string(bio), read_string(bio))
            elif op_ == 0x04:
                _ = bio.seek(8, 1)
                (key,) = deserialize_uint16(bio.read(2))
                schema_name = schemas.get(
                    deserialize_uint16(bio.read(2))[0],
                    Schema(0, '__schemaless__', 'cdr', ''),
                ).name
                channels[key] = Channel(
                    key,
                    schema_name,
                    read_string(bio),
                    read_string(bio),
                    read_bytes(bio),
                )
            elif op_ == 0x05:
                (size,) = deserialize_uint64(bio.read(8))
                (cid,) = deserialize_uint16(bio.read(2))
                _ = bio.seek(4, 1)
                (timestamp,) = deserialize_uint64(bio.read(8))
                msgcount += 1
                start_time = min(timestamp, start_time)
                end_time = max(timestamp, end_time)
                cstats[cid] += 1
                _ = bio.seek(size - 14, 1)
            elif op_ == 0x06:
                _ = bio.seek(8, 1)
                _, _, uncompressed_size, _ = deserialize_qqqi(bio.read(28))
                compression = read_string(bio)
                (compressed_size,) = deserialize_uint64(bio.read(8))
                bio = BytesIO(
                    DECOMPRESSORS[compression](bio.read(compressed_size), uncompressed_size),
                )
                bio_size = uncompressed_size
                nchunks += 1
            else:
                skip_sized(bio)

            if bio.tell() == bio_size and bio != self.bio:
                bio = self.bio
                bio_size = self.data_end

        self.statistics = Statistics(
            msgcount,
            len(schemas),
            len(channels),
            0,
            0,
            nchunks,
            start_time,
            end_time,
            cstats,
        )

    def messages_scan(
        self,
        connections: Iterable[Connection],
        start: int | None = None,
        stop: int | None = None,
    ) -> Generator[tuple[Connection, int, bytes], None, None]:
        """Read messages by scanning whole bag."""
        assert self.bio
        bio = self.bio
        bio_size = self.data_end
        _ = bio.seek(self.data_start)

        cmap = {x.id: x for x in connections}

        if start is None:
            start = 0
        if stop is None:
            stop = MAXSIZE

        while bio.tell() < bio_size:
            op_ = ord(bio.read(1))

            if op_ == 0x05:
                size, channel_id, _, timestamp, _ = deserialize_qhiqq(bio.read(30))
                data = bio.read(size - 22)
                if start <= timestamp < stop and channel_id in cmap:
                    yield cmap[channel_id], timestamp, data
            elif op_ == 0x06:
                (size,) = deserialize_uint64(bio.read(8))
                start_time, end_time, uncompressed_size, _ = deserialize_qqqi(bio.read(28))
                if start < end_time and start_time < stop:
                    compression = read_string(bio)
                    (compressed_size,) = deserialize_uint64(bio.read(8))
                    bio = BytesIO(
                        DECOMPRESSORS[compression](bio.read(compressed_size), uncompressed_size),
                    )
                    bio_size = uncompressed_size
                else:
                    _ = bio.seek(size - 28, 1)
            else:
                skip_sized(bio)

            if bio.tell() == bio_size and bio != self.bio:
                bio = self.bio
                bio_size = self.data_end

    def messages(
        self,
        connections: Iterable[Connection],
        start: int | None = None,
        stop: int | None = None,
    ) -> Generator[tuple[Connection, int, bytes], None, None]:
        """Read messages from bag.

        Args:
            connections: Iterable with connections to filter for.
            start: Yield only messages at or after this timestamp (ns).
            stop: Yield only messages before this timestamp (ns).

        Yields:
            tuples of connection, timestamp (ns), and rawdata.

        """
        assert self.bio

        if not self.chunks:
            yield from self.messages_scan(connections, start, stop)
            return

        channel_map = {  # pragma: no branch
            cid: conn
            for conn in connections
            if (
                cid := next(
                    (
                        cid
                        for cid, x in self.channels.items()
                        if x.schema == conn.msgtype and x.topic == conn.topic
                    ),
                    None,
                )
            )
        }

        chunks = [
            msgsrc(
                x,
                channel_map,
                start or x.message_start_time,
                stop or x.message_end_time + 1,
                self.bio,
            )
            for x in self.chunks
            if (start is None or start < x.message_end_time)
            and (stop is None or x.message_start_time < stop)
            and (any(x.channel_count.get(cid, 0) for cid in channel_map))
        ]

        for timestamp, offset, connection, data in heapq.merge(*chunks):
            if not offset:
                continue
            assert connection
            assert data
            yield connection, timestamp, data


def write_uint64(bio: BinaryIO, uint: int) -> None:
    """Serialize and write uint64."""
    bio.write(uint.to_bytes(8, byteorder='little'))


def write_uint32(bio: BinaryIO, uint: int) -> None:
    """Serialize and write uint32."""
    bio.write(uint.to_bytes(4, byteorder='little'))


def write_uint16(bio: BinaryIO, uint: int) -> None:
    """Serialize and write uint16."""
    bio.write(uint.to_bytes(2, byteorder='little'))


def write_string(bio: BinaryIO, string: str) -> None:
    """Serialize and write string."""
    data = string.encode()
    write_uint32(bio, len(data))
    bio.write(data)


def write_schema(bio: BinaryIO, schema: Schema) -> None:
    """Write schema."""
    rec = BytesIO()
    write_uint16(rec, schema.id)
    write_string(rec, schema.name)
    write_string(rec, schema.encoding)
    write_string(rec, schema.data)
    write_record(bio, 0x03, rec)


def write_channel(bio: BinaryIO, channel: Channel, schemas: Iterable[Schema]) -> None:
    """Write schema."""
    rec = BytesIO()
    write_uint16(rec, channel.id)
    write_uint16(rec, next(x.id for x in schemas if x.name == channel.schema))
    write_string(rec, channel.topic)
    write_string(rec, channel.message_encoding)
    write_uint32(rec, len(channel.metadata))
    rec.write(channel.metadata)
    write_record(bio, 0x04, rec)


def write_record(bio: BinaryIO, op_: int, record: BytesIO) -> None:
    """Write record."""
    bio.write(op_.to_bytes(1, byteorder='little'))
    write_uint64(bio, record.tell())
    bio.write(record.getbuffer())


@dataclass
class PendingChunk:
    """Chunk."""

    message_start_time: int
    message_end_time: int
    bio: BytesIO
    msgs: dict[int, list[tuple[int, int]]]


class McapWriter:
    """Mcap Storage Writer."""

    def __init__(self, path: Path, compression: CompressionMode) -> None:
        """Initialize sqlite3 storage."""
        self.path = path / f'{path.name}.mcap'
        self.bio = self.path.open('xb')

        _ = self.bio.write(b'\x89MCAP\x30\r\n')
        rec = BytesIO()
        write_string(rec, 'ros2')
        write_string(rec, f'rosbags-{version("rosbags")}')
        write_record(self.bio, 0x01, rec)

        self.schemas: list[Schema] = []
        self.channels: list[Channel] = []
        self.chunks: list[BytesIO] = []
        self.chunk = PendingChunk(2**63 - 1, 0, BytesIO(), defaultdict(list))

        self.compressor: Callable[[bytes], bytes]
        if compression == CompressionMode.STORAGE:
            self.compression = 'zstd'
            self.compressor = zstandard.ZstdCompressor().compress
        else:
            self.compression = ''
            self.compressor = lambda x: x

        self.message_start_time = 2**63 - 1
        self.message_end_time = 0
        self.channel_stats: dict[int, int] = {}

    def add_msgtype(self, connection: Connection) -> None:
        """Add a msgtype.

        Args:
            connection: Connection.

        """
        self.schemas.append(
            Schema(
                len(self.schemas) + 1,
                connection.msgtype,
                'ros2msg' if connection.msgdef.format == MessageDefinitionFormat.MSG else 'ros2idl',
                connection.msgdef.data,
            )
        )
        write_schema(self.chunk.bio, self.schemas[-1])

    def add_connection(self, connection: Connection, offered_qos_profiles: str) -> None:
        """Add a connection.

        Args:
            connection: Connection.
            offered_qos_profiles: Serialized QoS profiles.

        """
        metadata = BytesIO()
        write_string(metadata, 'offered_qos_profiles')
        write_string(metadata, offered_qos_profiles)

        self.channels.append(
            Channel(
                connection.id,
                connection.msgtype,
                connection.topic,
                cast('ConnectionExtRosbag2', connection.ext).serialization_format,
                metadata.getvalue(),
            )
        )
        self.channel_stats[connection.id] = 0
        write_channel(self.chunk.bio, self.channels[-1], self.schemas)

    def close_chunk(self) -> None:
        """Close pending chunk."""
        rec = BytesIO()
        write_uint64(rec, self.chunk.message_start_time)
        write_uint64(rec, self.chunk.message_end_time)
        write_uint64(rec, self.chunk.bio.tell())
        write_uint32(rec, 0)

        compressed = self.compressor(self.chunk.bio.getvalue())
        write_string(rec, self.compression)
        write_uint64(rec, len(compressed))
        rec.write(compressed)

        chunk_start = self.bio.tell()
        write_record(self.bio, 0x06, rec)
        chunk_end = self.bio.tell()

        offsets: list[tuple[int, int]] = []
        for cid, msgs in self.chunk.msgs.items():
            offsets.append((cid, self.bio.tell()))
            rec = BytesIO()
            write_uint16(rec, cid)
            write_uint32(rec, len(msgs) * 16)
            for ts, offset in msgs:
                write_uint64(rec, ts)
                write_uint64(rec, offset)
            write_record(self.bio, 0x07, rec)

        rec = BytesIO()
        write_uint64(rec, self.chunk.message_start_time)
        write_uint64(rec, self.chunk.message_end_time)
        write_uint64(rec, chunk_start)
        write_uint64(rec, chunk_end - chunk_start)
        write_uint32(rec, len(offsets) * 10)
        for cid, offset in offsets:
            write_uint16(rec, cid)
            write_uint64(rec, offset)
        write_uint64(rec, self.bio.tell() - chunk_end)
        write_string(rec, self.compression)
        write_uint64(rec, len(compressed))
        write_uint64(rec, self.chunk.bio.tell())
        self.chunks.append(rec)

        self.chunk = PendingChunk(2**63 - 1, 0, BytesIO(), defaultdict(list))

    def write(self, connection: Connection, timestamp: int, data: bytes | memoryview) -> None:
        """Write message to rosbag2.

        Args:
            connection: Connection to write message to.
            timestamp: Message timestamp (ns).
            data: Serialized message data.

        """
        self.message_start_time = min(timestamp, self.message_start_time)
        self.message_end_time = max(timestamp, self.message_end_time)
        self.channel_stats[connection.id] += 1
        self.chunk.message_start_time = min(timestamp, self.chunk.message_start_time)
        self.chunk.message_end_time = max(timestamp, self.chunk.message_end_time)
        self.chunk.msgs[connection.id].append((timestamp, self.chunk.bio.tell()))

        rec = BytesIO()
        write_uint16(rec, connection.id)
        write_uint32(rec, 0)
        write_uint64(rec, timestamp)
        write_uint64(rec, timestamp)
        rec.write(data)
        write_record(self.chunk.bio, 0x05, rec)

        if self.chunk.bio.tell() > 2**20:
            self.close_chunk()

    def close(self, version: int, metadata: str) -> None:
        """Close rosbag2 after writing.

        Closes open database transactions and writes metadata.yaml.

        """
        _ = version
        if self.chunk.bio.tell():
            self.close_chunk()

        metadata_start = self.bio.tell()
        rec = BytesIO()
        write_string(rec, 'rosbag2')
        sub = BytesIO()
        write_string(sub, 'serialized_metadata')
        write_string(sub, metadata)
        write_uint32(rec, sub.tell())
        rec.write(sub.getbuffer())
        write_record(self.bio, 0x0C, rec)
        metadata_end = self.bio.tell()

        rec = BytesIO()
        write_uint32(rec, 0)
        write_record(self.bio, 0x0F, rec)

        schema_start = self.bio.tell()
        for schema in self.schemas:
            write_schema(self.bio, schema)

        channel_start = self.bio.tell()
        for channel in self.channels:
            write_channel(self.bio, channel, self.schemas)

        chunk_start = self.bio.tell()
        for chunk in self.chunks:
            write_record(self.bio, 0x08, chunk)

        metadata_index_start = self.bio.tell()
        rec = BytesIO()
        write_uint64(rec, metadata_start)
        write_uint64(rec, metadata_end - metadata_start)
        write_string(rec, 'rosbag2')
        write_record(self.bio, 0x0D, rec)

        statistics_offset_start = self.bio.tell()
        rec = BytesIO()
        write_uint64(rec, sum(self.channel_stats.values()))
        write_uint16(rec, len(self.schemas))
        write_uint32(rec, len(self.channels))
        write_uint32(rec, 0)  # attachments
        write_uint32(rec, 1)  # metadata
        write_uint32(rec, len(self.chunks))
        write_uint64(rec, self.message_start_time)
        write_uint64(rec, self.message_end_time)
        write_uint32(rec, len(self.channel_stats) * 10)
        for cid, count in sorted(self.channel_stats.items()):
            write_uint16(rec, cid)
            write_uint64(rec, count)
        write_record(self.bio, 0x0B, rec)

        summary_offset_start = self.bio.tell()

        if schema_start != channel_start:
            rec = BytesIO()
            rec.write(b'\x03')
            write_uint64(rec, schema_start)
            write_uint64(rec, channel_start - schema_start)
            write_record(self.bio, 0x0E, rec)

        if channel_start != chunk_start:
            rec = BytesIO()
            rec.write(b'\x04')
            write_uint64(rec, channel_start)
            write_uint64(rec, chunk_start - channel_start)
            write_record(self.bio, 0x0E, rec)

        if chunk_start != metadata_index_start:
            rec = BytesIO()
            rec.write(b'\x08')
            write_uint64(rec, chunk_start)
            write_uint64(rec, metadata_index_start - chunk_start)
            write_record(self.bio, 0x0E, rec)

        rec = BytesIO()
        rec.write(b'\x0d')
        write_uint64(rec, metadata_index_start)
        write_uint64(rec, statistics_offset_start - metadata_index_start)
        write_record(self.bio, 0x0E, rec)

        rec = BytesIO()
        rec.write(b'\x0b')
        write_uint64(rec, statistics_offset_start)
        write_uint64(rec, summary_offset_start - statistics_offset_start)
        write_record(self.bio, 0x0E, rec)

        rec = BytesIO()
        write_uint64(rec, schema_start)
        write_uint64(rec, summary_offset_start)
        write_uint32(rec, 0)
        write_record(self.bio, 0x02, rec)
        _ = self.bio.write(b'\x89MCAP\x30\r\n')

        self.bio.close()



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Reference CDR Message Serializer and Deserializer."""

from __future__ import annotations

import sys
from struct import Struct, pack_into, unpack_from
from typing import TYPE_CHECKING, cast

import numpy as np

from rosbags.interfaces import Nodetype
from rosbags.serde import SerdeError
from rosbags.serde.utils import SIZEMAP

if TYPE_CHECKING:
    from typing import TypeAlias

    from numpy.typing import NDArray

    from rosbags.interfaces import Msgdef, Typestore
    from rosbags.interfaces.typing import Basename, FieldDesc
    from rosbags.typesys.store import Msgarg

    Array: TypeAlias = 'list[object] | list[str] | NDArray[np.float64]'
    BasetypeMap: TypeAlias = 'dict[Basename, Struct]'

BASETYPEMAP_LE: BasetypeMap = {
    'bool': Struct('?'),
    'int8': Struct('b'),
    'int16': Struct('<h'),
    'int32': Struct('<i'),
    'int64': Struct('<q'),
    'uint8': Struct('B'),
    'uint16': Struct('<H'),
    'uint32': Struct('<I'),
    'uint64': Struct('<Q'),
    'float32': Struct('<f'),
    'float64': Struct('<d'),
}

BASETYPEMAP_BE: BasetypeMap = {
    'bool': Struct('?'),
    'int8': Struct('b'),
    'int16': Struct('>h'),
    'int32': Struct('>i'),
    'int64': Struct('>q'),
    'uint8': Struct('B'),
    'uint16': Struct('>H'),
    'uint32': Struct('>I'),
    'uint64': Struct('>Q'),
    'float32': Struct('>f'),
    'float64': Struct('>d'),
}


def deserialize_number(
    rawdata: bytes,
    bmap: BasetypeMap,
    pos: int,
    basetype: Basename,
) -> tuple[bool | float | int, int]:
    """Deserialize a single boolean, float, or int.

    Args:
        rawdata: Serialized data.
        bmap: Basetype metadata.
        pos: Read position.
        basetype: Number type string.

    Returns:
        Deserialized number and new read position.

    """
    dtype, size = bmap[basetype], SIZEMAP[basetype]
    pos = (pos + size - 1) & -size
    return dtype.unpack_from(rawdata, pos)[0], pos + size


def deserialize_string(rawdata: bytes, bmap: BasetypeMap, pos: int) -> tuple[str, int]:
    """Deserialize a string value.

    Args:
        rawdata: Serialized data.
        bmap: Basetype metadata.
        pos: Read position.

    Returns:
        Deserialized string and new read position.

    """
    pos = (pos + 4 - 1) & -4
    length: int = bmap['int32'].unpack_from(rawdata, pos)[0]
    val = bytes(rawdata[pos + 4 : pos + 4 + length - 1])
    return val.decode(), pos + 4 + length


def deserialize_array(
    rawdata: bytes,
    bmap: BasetypeMap,
    pos: int,
    num: int,
    desc: FieldDesc,
    typestore: Typestore,
) -> tuple[Array, int]:
    """Deserialize an array of items of same type.

    Args:
        rawdata: Serialized data.
        bmap: Basetype metadata.
        pos: Read position.
        num: Number of elements.
        desc: Element type descriptor.
        typestore: Typestore.

    Returns:
        Deserialized array and new read position.

    Raises:
        SerdeError: Unexpected element type.

    """
    if desc[0] == Nodetype.BASE:
        if desc[1][0] == 'string':
            strs: list[str] = []
            while (num := num - 1) >= 0:
                val, pos = deserialize_string(rawdata, bmap, pos)
                strs.append(val)
            return strs, pos

        size = SIZEMAP[desc[1][0]]
        pos = (pos + size - 1) & -size
        ndarr = np.frombuffer(rawdata, dtype=desc[1][0], count=num, offset=pos)
        if (bmap is BASETYPEMAP_LE) != (sys.byteorder == 'little'):
            ndarr = ndarr.byteswap()  # no inplace on readonly array
        return ndarr, pos + num * SIZEMAP[desc[1][0]]

    if desc[0] == Nodetype.NAME:
        msgs: list[object] = []
        while (num := num - 1) >= 0:
            rosmsg, pos = deserialize_message(
                rawdata, bmap, pos, typestore.get_msgdef(desc[1]), typestore
            )
            msgs.append(rosmsg)
        return msgs, pos

    msg = f'Nested arrays {desc!r} are not supported.'
    raise SerdeError(msg)


def deserialize_message(
    rawdata: bytes,
    bmap: BasetypeMap,
    pos: int,
    msgdef: Msgdef[object],
    typestore: Typestore,
) -> tuple[object, int]:
    """Deserialize a message.

    Args:
        rawdata: Serialized data.
        bmap: Basetype metadata.
        pos: Read position.
        msgdef: Message definition.
        typestore: Typestore.

    Returns:
        Deserialized message and new read position.

    """
    values: list[object] = []

    for _, desc in msgdef.fields:
        if desc[0] == Nodetype.NAME:
            obj, pos = deserialize_message(
                rawdata, bmap, pos, typestore.get_msgdef(desc[1]), typestore
            )
            values.append(obj)

        elif desc[0] == Nodetype.BASE:
            if desc[1][0] == 'string':
                val, pos = deserialize_string(rawdata, bmap, pos)
                values.append(val)
            else:
                num, pos = deserialize_number(rawdata, bmap, pos, desc[1][0])
                values.append(num)

        elif desc[0] == Nodetype.ARRAY:
            subdesc, length = desc[1]
            arr, pos = deserialize_array(rawdata, bmap, pos, length, subdesc, typestore)
            values.append(arr)

        elif desc[0] == Nodetype.SEQUENCE:
            size, pos = deserialize_number(rawdata, bmap, pos, 'int32')
            arr, pos = deserialize_array(rawdata, bmap, pos, int(size), desc[1][0], typestore)
            values.append(arr)

    return msgdef.cls(*values), pos


def deserialize(rawdata: bytes, typename: str, typestore: Typestore) -> object:
    """Deserialize raw data into a message object.

    Args:
        rawdata: Serialized data.
        typename: Type to deserialize.
        typestore: Typestore,

    Returns:
        Deserialized message object.

    """
    _, little_endian = cast('tuple[int, int]', unpack_from('BB', rawdata, 0))

    msgdef = typestore.get_msgdef(typename)
    obj, _ = deserialize_message(
        rawdata[4:],
        BASETYPEMAP_LE if little_endian else BASETYPEMAP_BE,
        0,
        msgdef,
        typestore,
    )

    return obj


def serialize_number(
    rawdata: memoryview,
    bmap: BasetypeMap,
    pos: int,
    basetype: Basename,
    val: float,
) -> int:
    """Serialize a single boolean, float, or int.

    Args:
        rawdata: Serialized data.
        bmap: Basetype metadata.
        pos: Write position.
        basetype: Number type string.
        val: Value to serialize.

    Returns:
        Next write position.

    """
    dtype, size = bmap[basetype], SIZEMAP[basetype]
    pos = (pos + size - 1) & -size
    dtype.pack_into(rawdata, pos, val)
    return pos + size


def serialize_string(rawdata: memoryview, bmap: BasetypeMap, pos: int, val: str) -> int:
    """Deserialize a string value.

    Args:
        rawdata: Serialized data.
        bmap: Basetype metadata.
        pos: Write position.
        val: Value to serialize.

    Returns:
        Next write position.

    """
    bval = memoryview(val.encode())
    length = len(bval) + 1

    pos = (pos + 4 - 1) & -4
    bmap['int32'].pack_into(rawdata, pos, length)
    rawdata[pos + 4 : pos + 4 + length - 1] = bval
    return pos + 4 + length


def serialize_array(
    rawdata: memoryview,
    bmap: BasetypeMap,
    pos: int,
    desc: FieldDesc,
    val: Array,
    typestore: Typestore,
) -> int:
    """Serialize an array of items of same type.

    Args:
        rawdata: Serialized data.
        bmap: Basetype metadata.
        pos: Write position.
        desc: Element type descriptor.
        val: Value to serialize.
        typestore: Typestore.

    Returns:
        Next write position.

    Raises:
        SerdeError: Unexpected element type.

    """
    if desc[0] == Nodetype.BASE:
        if desc[1][0] == 'string':
            for item in val:
                pos = serialize_string(rawdata, bmap, pos, cast('str', item))
            return pos

        size = SIZEMAP[desc[1][0]]
        pos = (pos + size - 1) & -size
        size *= len(val)
        val = cast('NDArray[np.float64]', val)
        if (bmap is BASETYPEMAP_LE) != (sys.byteorder == 'little'):
            val = val.byteswap()  # no inplace on readonly array
        rawdata[pos : pos + size] = memoryview(val.tobytes())
        return pos + size

    if desc[0] == Nodetype.NAME:
        for item in val:
            pos = serialize_message(
                rawdata, bmap, pos, item, typestore.get_msgdef(desc[1]), typestore
            )
        return pos

    msg = f'Nested arrays {desc!r} are not supported.'
    raise SerdeError(msg)  # pragma: no cover


def serialize_message(
    rawdata: memoryview,
    bmap: BasetypeMap,
    pos: int,
    message: object,
    msgdef: Msgdef[object],
    typestore: Typestore,
) -> int:
    """Serialize a message.

    Args:
        rawdata: Serialized data.
        bmap: Basetype metadata.
        pos: Write position.
        message: Message object.
        msgdef: Message definition.
        typestore: Typestore.

    Returns:
        Next write position.

    """
    for fieldname, desc in msgdef.fields:
        val: Msgarg = getattr(message, fieldname)
        if desc[0] == Nodetype.NAME:
            pos = serialize_message(
                rawdata, bmap, pos, val, typestore.get_msgdef(desc[1]), typestore
            )

        elif desc[0] == Nodetype.BASE:
            if desc[1][0] == 'string':
                assert isinstance(val, str)
                pos = serialize_string(rawdata, bmap, pos, val)
            else:
                assert isinstance(val, int | float)
                pos = serialize_number(rawdata, bmap, pos, desc[1][0], val)

        elif desc[0] == Nodetype.ARRAY:
            assert isinstance(val, list | np.ndarray)
            lval: list[str] = cast('list[str]', val)
            pos = serialize_array(rawdata, bmap, pos, desc[1][0], lval, typestore)

        elif desc[0] == Nodetype.SEQUENCE:
            assert isinstance(val, list | np.ndarray)
            lval = cast('list[str]', val)
            size = len(lval)
            pos = serialize_number(rawdata, bmap, pos, 'int32', size)
            pos = serialize_array(rawdata, bmap, pos, desc[1][0], lval, typestore)

    return pos


def get_array_size(desc: FieldDesc, val: Array, size: int, typestore: Typestore) -> int:
    """Calculate size of an array.

    Args:
        desc: Element type descriptor.
        val: Array to calculate size of.
        size: Current size of message.
        typestore: Typestore.

    Returns:
        Size of val in bytes.

    Raises:
        SerdeError: Unexpected element type.

    """
    if desc[0] == Nodetype.BASE:
        if desc[1][0] == 'string':
            for item in val:
                size = (size + 4 - 1) & -4
                assert isinstance(item, str)
                size += 4 + len(item) + 1
            return size

        isize = SIZEMAP[desc[1][0]]
        size = (size + isize - 1) & -isize
        return size + isize * len(val)

    if desc[0] == Nodetype.NAME:
        for item in val:
            size = get_size(item, typestore.get_msgdef(desc[1]), typestore, size)
        return size

    msg = f'Nested arrays {desc!r} are not supported.'
    raise SerdeError(msg)  # pragma: no cover


def get_size(message: object, msgdef: Msgdef[object], typestore: Typestore, size: int = 0) -> int:
    """Calculate size of serialized message.

    Args:
        message: Message object.
        msgdef: Message definition.
        typestore: Typestore.
        size: Current size of message.

    Returns:
        Size of message in bytes.

    Raises:
        SerdeError: Unexpected array length in message.

    """
    for fieldname, desc in msgdef.fields:
        val: Msgarg = getattr(message, fieldname)
        if desc[0] == Nodetype.NAME:
            size = get_size(val, typestore.get_msgdef(desc[1]), typestore, size)

        elif desc[0] == Nodetype.BASE:
            if desc[1][0] == 'string':
                assert isinstance(val, str)
                size = (size + 4 - 1) & -4
                size += 4 + len(val.encode()) + 1
            else:
                isize = SIZEMAP[desc[1][0]]
                size = (size + isize - 1) & -isize
                size += isize

        elif desc[0] == Nodetype.ARRAY:
            subdesc, length = desc[1]
            assert isinstance(val, list | np.ndarray)
            lval: list[str] = cast('list[str]', val)
            if len(lval) != length:
                msg = f'Unexpected array length: {len(lval)} != {length}.'
                raise SerdeError(msg)
            size = get_array_size(subdesc, lval, size, typestore)

        elif desc[0] == Nodetype.SEQUENCE:
            assert isinstance(val, list | np.ndarray)
            lval = cast('list[str]', val)
            size = (size + 4 - 1) & -4
            size += 4
            size = get_array_size(desc[1][0], lval, size, typestore)

    return size


def serialize(
    message: object,
    typename: str,
    typestore: Typestore,
    *,
    little_endian: bool = sys.byteorder == 'little',
) -> memoryview:
    """Serialize message object to bytes.

    Args:
        message: Message object.
        typename: Type to serialize.
        little_endian: Should use little endianness.
        typestore: Typestore.

    Returns:
        Serialized bytes.

    """
    msgdef = typestore.get_msgdef(typename)
    size = 4 + get_size(message, msgdef, typestore)
    rawdata = memoryview(bytearray(size))

    pack_into('BB', rawdata, 0, 0, little_endian)
    pos = serialize_message(
        rawdata[4:],
        BASETYPEMAP_LE if little_endian else BASETYPEMAP_BE,
        0,
        message,
        msgdef,
        typestore,
    )
    assert pos + 4 == size
    return rawdata.toreadonly()



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Reader Tests."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, cast
from unittest.mock import MagicMock, patch

import pytest
import zstandard
from ruamel.yaml import YAML

from rosbags.interfaces import (
    Connection,
    ConnectionExtRosbag2,
    MessageDefinition,
    MessageDefinitionFormat,
)
from rosbags.rosbag2 import Reader, ReaderError
from rosbags.rosbag2.metadata import Metadata
from rosbags.rosbag2.reader import DirectoryReader

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable


METADATA = """
rosbag2_bagfile_information:
  version: 4
  storage_identifier: mock
  relative_file_paths:
    - db0.dat{extension}
    - db1.dat{extension}
  duration:
    nanoseconds: 42
  starting_time:
    nanoseconds_since_epoch: 666
  message_count: 4
  topics_with_message_count:
    - topic_metadata:
        name: /poly
        type: geometry_msgs/msg/Polygon
        serialization_format: cdr
        offered_qos_profiles: ""
      message_count: 1
    - topic_metadata:
        name: /magn
        type: sensor_msgs/msg/MagneticField
        serialization_format: cdr
        offered_qos_profiles: ""
      message_count: 2
    - topic_metadata:
        name: /joint
        type: trajectory_msgs/msg/JointTrajectory
        serialization_format: cdr
        offered_qos_profiles: ""
      message_count: 1
  compression_format: {compression_format}
  compression_mode: {compression_mode}
"""

METADATA_EMPTY = """
rosbag2_bagfile_information:
  version: 8
  storage_identifier: mock
  relative_file_paths:
    - db0.dat
    - db1.dat
  duration:
    nanoseconds: 0
  starting_time:
    nanoseconds_since_epoch: 0
  message_count: 0
  topics_with_message_count: []
  compression_format: ""
  compression_mode: ""
  files:
  - duration:
      nanoseconds: 0
    message_count: 0
    path: db0.dat
    starting_time:
      nanoseconds_since_epoch: 0
  - duration:
      nanoseconds: 0
    message_count: 0
    path: db1.dat
    starting_time:
      nanoseconds_since_epoch: 0
  custom_data:
    key1: value1
    key2: value2
  ros_distro: rosbags
"""


@pytest.fixture(autouse=True)
def mock_storage() -> Generator[MagicMock, None, None]:
    """Inject fake storage plugin."""
    mock_storage = MagicMock(name='MockStorage')

    with patch.dict(
        'rosbags.rosbag2.reader.DirectoryReader.STORAGE_PLUGINS', {'mock': mock_storage}
    ):
        yield mock_storage


@pytest.fixture
def empty_bag(tmp_path: Path) -> Path:
    """Manually construct empty bag."""
    _ = (tmp_path / 'metadata.yaml').write_text(METADATA_EMPTY)
    (tmp_path / 'db0.dat').touch()
    (tmp_path / 'db1.dat').touch()
    return tmp_path


@pytest.fixture
def nonempty_bag(tmp_path: Path) -> Path:
    """Manually construct empty bag."""
    _ = (tmp_path / 'metadata.yaml').write_text(
        METADATA.format(extension='', compression_format='""', compression_mode='""'),
    )
    (tmp_path / 'db0.dat').touch()
    (tmp_path / 'db1.dat').touch()
    return tmp_path


@pytest.fixture(params=['none', 'file', 'message'])
def bag_with_compression(
    request: pytest.FixtureRequest,
    tmp_path: Path,
    mock_storage: MagicMock,
) -> Path:
    """Manually construct bag with compression."""
    param: str = request.param
    _ = (tmp_path / 'metadata.yaml').write_text(
        METADATA.format(
            extension='' if param != 'file' else '.zstd',
            compression_format='""' if param == 'none' else 'zstd',
            compression_mode='""' if param == 'none' else param.upper(),
        ),
    )

    comp = zstandard.ZstdCompressor()
    textcomp = comp.compress if request.param == 'message' else bytes

    db0 = tmp_path / 'db0.dat'
    db0.write_bytes(b'mockdata0')

    db1 = tmp_path / 'db1.dat'
    db1.write_bytes(b'mockdata1')

    if param == 'file':
        for item in [db0, db1]:
            with item.open('rb') as ifh, item.with_suffix('.dat.zstd').open('wb') as ofh:
                _ = comp.copy_stream(ifh, ofh)
                item.unlink()

    defargs = (
        MessageDefinition(MessageDefinitionFormat.NONE, ''),
        '',
        0,
        ConnectionExtRosbag2('cdr', []),
        None,
    )

    class MockStorage:
        """Mock Storage."""

        def __init__(self, path: Path) -> None:
            """Initialize."""
            self.index = int(path.stem[-1])
            self.path = path
            self.connections = (
                [
                    Connection(201, '/poly', 'geometry_msgs/msg/Polygon', *defargs),
                    Connection(101, '/magn', 'sensor_msgs/msg/MagneticField', *defargs),
                    Connection(301, '/joint', 'trajectory_msgs/msg/JointTrajectory', *defargs),
                ]
                if self.index
                else []
            )

        def open(self) -> None:
            """Open storage."""
            assert self.path.read_bytes().startswith(f'mockdata{self.index}'.encode())

        def close(self) -> None:
            """Close storage."""

        def messages(
            self,
            connections: Iterable[Connection],
            start: int | None = None,
            stop: int | None = None,
        ) -> Generator[tuple[Connection, int, bytes], None, None]:
            """Messages."""
            if not self.index:
                return

            cids = [x.id for x in connections]
            start = start or 0
            stop = stop or 1000

            for cid, timestamp, data in (
                (201, 666, textcomp(b'poly message')),
                (101, 708, textcomp(b'magn message')),
                (101, 708, textcomp(b'magn message')),
                (301, 708, textcomp(b'joint message')),
            ):
                if cid in cids and start <= timestamp < stop:
                    yield next(x for x in self.connections if x.id == cid), timestamp, data

    mock_storage.side_effect = MockStorage

    return tmp_path


def test_empty_metadata_conforms_to_spec() -> None:
    """Test empty metadata conforms to spec."""
    yaml = YAML(typ='safe')
    dct = cast(
        'dict[str, Metadata]',
        yaml.load(METADATA_EMPTY),
    )
    assert dct['rosbag2_bagfile_information'].keys() == Metadata.__annotations__.keys()


def test_directory_reader_raises_on_broken_fs_layouts(tmp_path: Path) -> None:
    """Test directory reader raises on broken filesystem layouts."""
    with pytest.raises(FileNotFoundError, match=r'metadata file .* does not exist'):
        DirectoryReader(tmp_path)

    metadata = tmp_path / 'metadata.yaml'
    metadata.touch()

    with (
        pytest.raises(PermissionError),
        patch.object(Path, 'read_text', side_effect=PermissionError),
    ):
        DirectoryReader(tmp_path).open()

    with (
        pytest.raises(ReaderError, match='not read'),
        patch.object(Path, 'read_text', side_effect=IsADirectoryError),
    ):
        DirectoryReader(tmp_path).open()

    _ = metadata.write_text('  invalid:\nthis is not yaml')
    with pytest.raises(ReaderError, match='not load YAML from'):
        DirectoryReader(tmp_path).open()

    _ = metadata.write_text('foo:')
    with pytest.raises(ReaderError, match='key is missing'):
        DirectoryReader(tmp_path).open()

    _ = metadata.write_text(METADATA_EMPTY)
    with pytest.raises(ReaderError, match='files are missing'):
        DirectoryReader(tmp_path).open()


def test_directory_reader_raises_on_unsupported_features(tmp_path: Path) -> None:
    """Test directory reader raises if fs layout is broken."""
    metadata = tmp_path / 'metadata.yaml'
    (tmp_path / 'db0.dat').touch()
    (tmp_path / 'db1.dat').touch()

    _ = metadata.write_text(
        METADATA.format(
            extension='',
            compression_format='""',
            compression_mode='""',
        ).replace('version: 4', 'version: 999'),
    )
    with pytest.raises(ReaderError, match='version 999'):
        DirectoryReader(tmp_path).open()

    _ = metadata.write_text(
        METADATA.format(
            extension='',
            compression_format='""',
            compression_mode='""',
        ).replace('mock', 'hdf5'),
    )
    with pytest.raises(ReaderError, match='Storage plugin'):
        DirectoryReader(tmp_path).open()

    _ = metadata.write_text(
        METADATA.format(
            extension='',
            compression_format='""',
            compression_mode='""',
        ).replace('cdr', 'bson'),
    )
    with pytest.raises(ReaderError, match='Serialization format'):
        DirectoryReader(tmp_path).open()

    _ = metadata.write_text(
        METADATA.format(
            extension='',
            compression_format='"gz"',
            compression_mode='"file"',
        ),
    )
    with pytest.raises(ReaderError, match='Compression format'):
        DirectoryReader(tmp_path).open()


def test_directory_reader_raises_on_storage_open_exception(
    empty_bag: Path,
    mock_storage: MagicMock,
) -> None:
    """Test directory reader raises on storage open exception."""

    class OpenError(Exception): ...

    first = MagicMock()
    second = MagicMock()
    second.open.side_effect = OpenError
    mock_storage.side_effect = [first, second]

    with pytest.raises(OpenError):
        DirectoryReader(empty_bag).open()

    first.open.assert_called_once()
    first.close.assert_called_once()
    second.open.assert_called_once()
    second.close.assert_not_called()


def test_directory_reader_has_correct_metadata_on_empty_bag(
    empty_bag: Path, mock_storage: MagicMock
) -> None:
    """Test directory reader has correct metadata on empty bag."""
    storage = MagicMock()
    mock_storage.return_value = storage

    reader = DirectoryReader(empty_bag)
    reader.open()
    metadata = reader.metadata
    assert metadata.message_count == 0
    assert metadata.start_time == 2**63 - 1
    assert metadata.end_time == 0
    assert metadata.duration == 0
    assert metadata.custom
    assert metadata.custom['key1'] == 'value1'
    assert metadata.custom['key2'] == 'value2'
    assert metadata.ros_distro == 'rosbags'


def test_directory_reader_has_correct_metadata_on_nonempty_bag(
    nonempty_bag: Path, mock_storage: MagicMock
) -> None:
    """Test directory reader has correct metadata on empty bag."""
    storage = MagicMock()
    mock_storage.return_value = storage

    reader = DirectoryReader(nonempty_bag)
    reader.open()
    metadata = reader.metadata
    assert metadata.message_count == 4
    assert metadata.start_time == 666
    assert metadata.end_time == 709
    assert metadata.duration == 43
    assert not metadata.custom
    assert not metadata.ros_distro


def test_directory_reader_integrates_msgdefs_into_connections(
    nonempty_bag: Path,
    mock_storage: MagicMock,
) -> None:
    """Test directory reader integrates msgdefs into connections."""
    storage = MagicMock()
    storage.connections = [
        Connection(
            1,
            '/poly',
            'geometry_msgs/msg/Polygon',
            MessageDefinition(MessageDefinitionFormat.MSG, 'uint8 foo'),
            '',
            0,
            ConnectionExtRosbag2('cdr', []),
            None,
        )
    ]
    mock_storage.return_value = storage

    reader = DirectoryReader(nonempty_bag)
    reader.open()
    assert reader.connections[0].msgdef.data == 'uint8 foo'


def test_reader_raises_on_missing_path(tmp_path: Path) -> None:
    """Test reader raises on missing path."""
    with pytest.raises(FileNotFoundError):
        _ = Reader(tmp_path / 'missing')


def test_reader_selects_plugins(tmp_path: Path) -> None:
    """Test reader raises on missing path."""
    mockdir = MagicMock()
    mockdb3 = MagicMock()
    with patch.dict(
        'rosbags.rosbag2.reader.Reader.STORAGE_PLUGINS', {'dir': mockdir, '.db3': mockdb3}
    ):
        _ = Reader(tmp_path)
        mockdir.assert_called_once_with(tmp_path)
        mockdb3.assert_not_called()
        mockdir.reset_mock()

        dbpath = tmp_path / 'foo.db3'
        dbpath.touch()
        _ = Reader(dbpath)
        mockdir.assert_not_called()
        mockdb3.assert_called_once_with(dbpath)

        dbpath = tmp_path / 'unknown.plug'
        dbpath.touch()
        with pytest.raises(ReaderError, match='Unrecognized storage format'):
            _ = Reader(dbpath)


def test_reader_passes_messages_to_plugins(bag_with_compression: Path) -> None:
    """Test reader passes messages to plugins."""
    with Reader(bag_with_compression) as reader:
        gen = reader.messages()

        cmap = {x.topic: x for x in reader.connections}
        assert next(gen) == (cmap['/poly'], 666, b'poly message')
        assert next(gen) == (cmap['/magn'], 708, b'magn message')
        assert next(gen) == (cmap['/magn'], 708, b'magn message')
        assert next(gen) == (cmap['/joint'], 708, b'joint message')

        with pytest.raises(StopIteration):
            _ = next(gen)

        magn_connections = [x for x in reader.connections if x.topic == '/magn']
        gen = reader.messages(connections=magn_connections)
        assert next(gen)[0] == cmap['/magn']
        assert next(gen)[0] == cmap['/magn']
        with pytest.raises(StopIteration):
            _ = next(gen)

        gen = reader.messages(reader.connections, start=667)
        assert next(gen)[0] == cmap['/magn']
        assert next(gen)[0] == cmap['/magn']
        assert next(gen)[0] == cmap['/joint']
        with pytest.raises(StopIteration):
            _ = next(gen)

        gen = reader.messages(reader.connections, stop=667)
        assert next(gen)[0] == cmap['/poly']
        with pytest.raises(StopIteration):
            _ = next(gen)

        gen = reader.messages(connections=magn_connections, stop=667)
        with pytest.raises(StopIteration):
            _ = next(gen)

        gen = reader.messages(reader.connections, start=666, stop=666)
        with pytest.raises(StopIteration):
            _ = next(gen)


def test_reader_raises_if_closed(nonempty_bag: Path) -> None:
    """Test reader raises if methods called while closed."""
    reader = Reader(nonempty_bag)

    def check() -> None:
        with pytest.raises(ReaderError, match='Rosbag is not open'):
            _ = reader.duration

        with pytest.raises(ReaderError, match='Rosbag is not open'):
            _ = reader.start_time

        with pytest.raises(ReaderError, match='Rosbag is not open'):
            _ = reader.end_time

        with pytest.raises(ReaderError, match='Rosbag is not open'):
            _ = reader.message_count

        with pytest.raises(ReaderError, match='Rosbag is not open'):
            _ = reader.compression_format

        with pytest.raises(ReaderError, match='Rosbag is not open'):
            _ = reader.compression_mode

        with pytest.raises(ReaderError, match='Rosbag is not open'):
            _ = reader.connections

        with pytest.raises(ReaderError, match='Rosbag is not open'):
            _ = reader.topics

        with pytest.raises(ReaderError, match='Rosbag is not open'):
            _ = reader.ros_distro

        with pytest.raises(ReaderError, match='Rosbag is not open'):
            _ = next(reader.messages())

    check()

    with reader:
        _ = reader.duration
        _ = reader.start_time
        _ = reader.end_time
        _ = reader.message_count
        _ = reader.compression_format
        _ = reader.compression_mode
        _ = reader.connections
        _ = reader.topics
        _ = reader.ros_distro
        with pytest.raises(StopIteration):
            _ = next(reader.messages())

    check()



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Sqlite3 Storage Tests."""

from __future__ import annotations

import sqlite3
from typing import TYPE_CHECKING

import pytest

from rosbags.rosbag2.errors import ReaderError
from rosbags.rosbag2.storage_sqlite3 import Sqlite3Reader

if TYPE_CHECKING:
    from pathlib import Path

SQLITE_SCHEMA_V1 = """
CREATE TABLE topics(
  id INTEGER PRIMARY KEY,
  name TEXT NOT NULL,
  type TEXT NOT NULL,
  serialization_format TEXT NOT NULL
);
CREATE TABLE messages(
  id INTEGER PRIMARY KEY,
  topic_id INTEGER NOT NULL,
  timestamp INTEGER NOT NULL,
  data BLOB NOT NULL
);
CREATE INDEX timestamp_idx ON messages (timestamp ASC);
"""

SQLITE_SCHEMA_V2 = """
CREATE TABLE topics(
  id INTEGER PRIMARY KEY,
  name TEXT NOT NULL,
  type TEXT NOT NULL,
  serialization_format TEXT NOT NULL,
  offered_qos_profiles TEXT NOT NULL
);
CREATE TABLE messages(
  id INTEGER PRIMARY KEY,
  topic_id INTEGER NOT NULL,
  timestamp INTEGER NOT NULL,
  data BLOB NOT NULL
);
CREATE INDEX timestamp_idx ON messages (timestamp ASC);
"""

SQLITE_SCHEMA_V3 = """
CREATE TABLE schema(
  schema_version INTEGER PRIMARY KEY,
  ros_distro TEXT NOT NULL
);
CREATE TABLE metadata(
  id INTEGER PRIMARY KEY,
  metadata_version INTEGER NOT NULL,
  metadata TEXT NOT NULL
);
CREATE TABLE topics(
  id INTEGER PRIMARY KEY,
  name TEXT NOT NULL,
  type TEXT NOT NULL,
  serialization_format TEXT NOT NULL,
  offered_qos_profiles TEXT NOT NULL
);
CREATE TABLE messages(
  id INTEGER PRIMARY KEY,
  topic_id INTEGER NOT NULL,
  timestamp INTEGER NOT NULL,
  data BLOB NOT NULL
);
CREATE INDEX timestamp_idx ON messages (timestamp ASC);
INSERT INTO schema(schema_version, ros_distro) VALUES (3, 'rosbags');
"""

SQLITE_SCHEMA_V4 = """
CREATE TABLE schema(
  schema_version INTEGER PRIMARY KEY,
  ros_distro TEXT NOT NULL
);
CREATE TABLE metadata(
  id INTEGER PRIMARY KEY,
  metadata_version INTEGER NOT NULL,
  metadata TEXT NOT NULL
);
CREATE TABLE topics(
  id INTEGER PRIMARY KEY,
  name TEXT NOT NULL,
  type TEXT NOT NULL,
  serialization_format TEXT NOT NULL,
  offered_qos_profiles TEXT NOT NULL,
  type_description_hash TEXT NOT NULL
);
CREATE TABLE message_definitions(
  id INTEGER PRIMARY KEY,
  topic_type TEXT NOT NULL,
  encoding TEXT NOT NULL,
  encoded_message_definition TEXT NOT NULL,
  type_description_hash TEXT NOT NULL
);
CREATE TABLE messages(
  id INTEGER PRIMARY KEY,
  topic_id INTEGER NOT NULL,
  timestamp INTEGER NOT NULL,
  data BLOB NOT NULL
);
CREATE INDEX timestamp_idx ON messages (timestamp ASC);
INSERT INTO schema(schema_version, ros_distro) VALUES (4, 'rosbags');
"""


@pytest.fixture
def database(tmp_path: Path) -> Path:
    """Create empty database."""
    path = tmp_path / 'db.db3'
    con = sqlite3.connect(path)
    con.close()

    return path


@pytest.fixture(
    params=[
        (1, SQLITE_SCHEMA_V1),
        (2, SQLITE_SCHEMA_V2),
        (3, SQLITE_SCHEMA_V3),
        (4, SQLITE_SCHEMA_V4),
    ],
)
def schema(request: pytest.FixtureRequest, tmp_path: Path) -> int:
    """Create empty database."""
    version: int
    version, sql = request.param

    path = tmp_path / 'db.db3'
    con = sqlite3.connect(path)
    with con:
        _ = con.executescript(sql)
    con.close()

    return version


@pytest.fixture
def add_connections(database: Path, schema: int) -> None:
    """Add connections to database."""
    con = sqlite3.connect(database)
    with con:
        if schema >= 4:
            _ = con.execute(
                (
                    'INSERT INTO message_definitions(topic_type, encoding,'
                    ' encoded_message_definition, type_description_hash) VALUES (?, ?, ?, ?);'
                ),
                (
                    'std_msgs/msg/Empty',
                    'ros2msg',
                    '',
                    'RIHS01_20b625256f32d5dbc0d04fee44f43c41e51c70d3502f84b4a08e7a9c26a96312',
                ),
            )
            _ = con.execute(
                (
                    'INSERT INTO topics('
                    'name, type, serialization_format, offered_qos_profiles, type_description_hash'
                    ') VALUES (?, ?, ?, ?, ?), (?, ?, ?, ?, ?)'
                ),
                ('t1', 'm1', 'cdr', '', 'h1', 't2', 'm2', 'cdr', '', 'h2'),
            )
    con.close()


@pytest.fixture
def add_messages(database: Path) -> None:
    """Add messages to database."""
    con = sqlite3.connect(database)
    with con:
        _ = con.execute(
            ('INSERT INTO messages(topic_id, timestamp, data)VALUES (?, ?, ?), (?, ?, ?);'),
            (1, 42, b'', 1, 666, b''),
        )
    con.close()


def test_open_raises_on_missing_tables(database: Path) -> None:
    """Test open raises on missing tables."""
    reader = Sqlite3Reader(database)
    with pytest.raises(ReaderError, match='missing tables'):
        reader.open()


def test_detects_schema_version(database: Path, schema: int) -> None:
    """Test schema version is detected."""
    reader = Sqlite3Reader(database)
    reader.open()
    assert reader.schema == schema
    reader.close()


@pytest.mark.usefixtures('schema')
@pytest.mark.parametrize('schema', [(4, SQLITE_SCHEMA_V4)], indirect=True)
def test_empty_database_has_correct_metadata(database: Path) -> None:
    """Test empty database has correct metadata."""
    reader = Sqlite3Reader(database)
    reader.open()
    assert reader.metadata.duration == 0
    assert reader.metadata.start_time == 2**63 - 1
    assert reader.metadata.end_time == 0
    assert reader.metadata.message_count == 0
    assert reader.connections == []
    reader.close()


@pytest.mark.usefixtures('schema', 'add_connections')
@pytest.mark.parametrize('schema', [(4, SQLITE_SCHEMA_V4)], indirect=True)
def test_database_with_empty_connections_has_correct_metadata(database: Path) -> None:
    """Test database with empty connections has correct metadata."""
    reader = Sqlite3Reader(database)
    reader.open()
    assert reader.metadata.duration == 0
    assert reader.metadata.start_time == 2**63 - 1
    assert reader.metadata.end_time == 0
    assert reader.metadata.message_count == 0
    assert len(reader.connections) == 2
    assert reader.connections[0].msgcount == 0
    assert reader.connections[1].msgcount == 0
    reader.close()


@pytest.mark.usefixtures('schema', 'add_connections', 'add_messages')
@pytest.mark.parametrize('schema', [(4, SQLITE_SCHEMA_V4)], indirect=True)
def test_database_with_messages_has_correct_metadata(database: Path) -> None:
    """Test database with messages has correct metadata."""
    reader = Sqlite3Reader(database)
    reader.open()
    assert reader.metadata.duration == 667 - 42
    assert reader.metadata.start_time == 42
    assert reader.metadata.end_time == 667
    assert reader.metadata.message_count == 2
    assert len(reader.connections) == 2
    assert reader.connections[0].msgcount == 2
    assert reader.connections[1].msgcount == 0
    reader.close()


@pytest.mark.usefixtures('add_connections')
def test_type_definitions_are_read(database: Path, schema: int) -> None:
    """Test type definitions are read."""
    reader = Sqlite3Reader(database)
    reader.open()
    assert (schema < 4) ^ bool(reader.msgtypes)
    reader.close()


@pytest.mark.parametrize('database', [(4, SQLITE_SCHEMA_V4)], indirect=True)
def test_messages_raises_on_closed_reader(database: Path) -> None:
    """Test messages raises on closed reader."""
    reader = Sqlite3Reader(database)
    with pytest.raises(AssertionError):
        _ = next(reader.messages([]))


@pytest.mark.usefixtures('schema', 'add_connections', 'add_messages')
@pytest.mark.parametrize('schema', [(4, SQLITE_SCHEMA_V4)], indirect=True)
def test_messages_yields_data(database: Path) -> None:
    """Test database with messages has correct metadata."""
    reader = Sqlite3Reader(database)
    reader.open()
    assert list(reader.messages(())) == []
    assert list(reader.messages(reader.connections, start=1000)) == []
    assert list(reader.messages(reader.connections, stop=40)) == []
    assert list(reader.messages(reader.connections, stop=100)) == [(reader.connections[0], 42, b'')]
    reader.close()


@pytest.mark.usefixtures(
    'schema',
)
@pytest.mark.parametrize('schema', [(4, SQLITE_SCHEMA_V4)], indirect=True)
def test_messages_fiters(database: Path) -> None:
    """Test message filters."""
    con = sqlite3.connect(database)

    with con:
        _ = con.execute(
            'INSERT INTO topics VALUES(?, ?, ?, ?, ?, ?)',
            (1, '/poly', 'geometry_msgs/msg/Polygon', 'cdr', '', ''),
        )
        _ = con.execute(
            'INSERT INTO topics VALUES(?, ?, ?, ?, ?, ?)',
            (2, '/magn', 'sensor_msgs/msg/MagneticField', 'cdr', '', ''),
        )
        _ = con.execute(
            'INSERT INTO topics VALUES(?, ?, ?, ?, ?, ?)',
            (3, '/joint', 'trajectory_msgs/msg/JointTrajectory', 'cdr', '', ''),
        )
        _ = con.execute(
            'INSERT INTO messages VALUES(?, ?, ?, ?)',
            (1, 1, 666, b'poly message'),
        )
        _ = con.execute(
            'INSERT INTO messages VALUES(?, ?, ?, ?)',
            (2, 2, 708, b'magn message'),
        )
        _ = con.execute(
            'INSERT INTO messages VALUES(?, ?, ?, ?)',
            (3, 2, 708, b'magn message'),
        )
        _ = con.execute(
            'INSERT INTO messages VALUES(?, ?, ?, ?)',
            (4, 3, 708, b'joint message'),
        )
    con.close()

    reader = Sqlite3Reader(database)
    reader.open()
    metadata = reader.metadata
    assert metadata.duration == 43
    assert metadata.start_time == 666
    assert metadata.end_time == 709
    assert metadata.message_count == 4
    assert [x.id for x in reader.connections] == [1, 2, 3]

    gen = reader.messages(reader.connections)

    cmap = {x.topic: x for x in reader.connections}
    assert next(gen) == (cmap['/poly'], 666, b'poly message')
    assert next(gen) == (cmap['/magn'], 708, b'magn message')
    assert next(gen) == (cmap['/magn'], 708, b'magn message')
    assert next(gen) == (cmap['/joint'], 708, b'joint message')

    with pytest.raises(StopIteration):
        _ = next(gen)

    magn_connections = [x for x in reader.connections if x.topic == '/magn']
    gen = reader.messages(connections=magn_connections)
    assert next(gen)[0] == cmap['/magn']
    assert next(gen)[0] == cmap['/magn']
    with pytest.raises(StopIteration):
        _ = next(gen)

    gen = reader.messages(reader.connections, start=667)
    assert next(gen)[0] == cmap['/magn']
    assert next(gen)[0] == cmap['/magn']
    assert next(gen)[0] == cmap['/joint']
    with pytest.raises(StopIteration):
        _ = next(gen)

    gen = reader.messages(reader.connections, stop=667)
    assert next(gen)[0] == cmap['/poly']
    with pytest.raises(StopIteration):
        _ = next(gen)

    gen = reader.messages(connections=magn_connections, stop=667)
    with pytest.raises(StopIteration):
        _ = next(gen)

    gen = reader.messages(reader.connections, start=666, stop=666)
    with pytest.raises(StopIteration):
        _ = next(gen)

    reader.close()



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Rosbags message serialization and deserialization.

Serializers and deserializers convert between python messages objects and
the common rosbag serialization formats. Computationally cheap functions
convert directly between different serialization formats.

"""

from .errors import SerdeError

__all__ = [
    'SerdeError',
]



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Rosbags support for rosbag1 files.

Readers and writers provide access to metadata and raw message content saved
in the rosbag1 format.

Supported versions:
  - Rosbag1 v2.0

"""

from .reader import Reader, ReaderError
from .writer import Writer, WriterError

__all__ = [
    'Reader',
    'ReaderError',
    'Writer',
    'WriterError',
]



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Type store."""

from __future__ import annotations

import json
import sys
from hashlib import md5, sha256
from importlib.util import module_from_spec, spec_from_loader
from itertools import starmap
from struct import pack_into
from typing import TYPE_CHECKING, Protocol, TypeVar, cast

from rosbags.interfaces import Msgdef, Nodetype
from rosbags.serde.cdr import generate_deserialize_cdr, generate_getsize_cdr, generate_serialize_cdr
from rosbags.serde.ros1 import (
    generate_cdr_to_ros1,
    generate_deserialize_ros1,
    generate_getsize_ros1,
    generate_ros1_to_cdr,
    generate_serialize_ros1,
)
from rosbags.typesys.base import TypesysError

from .codegen import generate_python_code
from .msg import denormalize_msgtype

if TYPE_CHECKING:
    from types import ModuleType
    from typing import ClassVar, TypeAlias, TypedDict

    from numpy.typing import ArrayLike

    from rosbags.interfaces import Bitcvt, BitcvtSize
    from rosbags.interfaces.typing import (
        FieldDesc,
        Typesdict,
    )

    class FieldType(TypedDict):
        """Field type."""

        type_id: int
        capacity: int
        string_capacity: int
        nested_type_name: str

    class Field(TypedDict):
        """Field."""

        name: str
        type: FieldType

    class Struct(TypedDict):
        """Struct."""

        type_name: str
        fields: list[Field]

    class MsgType(Protocol):
        """AnyInit."""

        __msgtype__: ClassVar[str]

        def __init__(self, *args: Msgarg, **kwargs: Msgarg) -> None:
            """Accept any args."""
            raise NotImplementedError

    class Msg(Protocol):
        """AnyInit."""

        __msgtype__: ClassVar[str]

    Msgarg: TypeAlias = Msg | list[Msg] | bool | int | float | str | list[str] | ArrayLike


T = TypeVar('T')

TIDMAP = {
    'int8': 2,
    'uint8': 3,
    'int16': 4,
    'uint16': 5,
    'int32': 6,
    'uint32': 7,
    'int64': 8,
    'uint64': 9,
    'float32': 10,
    'float64': 11,
    'float128': 12,
    'char': 13,
    # Unsupported 'wchar': 14,
    'bool': 15,
    'byte': 16,
    'string': 17,
    # Unsupported 'wstring': 18,
    # Unsupported 'fixed_string': 19,
    # Unsupported 'fixed_wstring': 20,
    'bounded_string': 21,
    # Unsupported 'bounded_wstring': 22,
}


class Typestore:
    """Type store."""

    types: dict[str, type[MsgType]]
    fielddefs: Typesdict

    def __init__(self, base: ModuleType | None = None) -> None:
        """Initialize."""
        self.cache: dict[str, Msgdef[object]] = {}
        self.types = {}
        self.fielddefs = {}
        if base:
            fielddefs: Typesdict = base.FIELDDEFS
            self.fielddefs.update(fielddefs)
            self.types.update({k: getattr(base, k.replace('/', '__')) for k in fielddefs})

    def deserialize_cdr(self, rawdata: bytes | memoryview, typename: str) -> object:
        """Deserialize raw data into a message object.

        Args:
            rawdata: Serialized data.
            typename: Message type name.

        Returns:
            Deserialized message object.

        """
        little_endian = bool(rawdata[1])

        msgdef = self.get_msgdef(typename)
        func = msgdef.deserialize_cdr_le if little_endian else msgdef.deserialize_cdr_be
        message, pos = func(rawdata[4:], 0, msgdef.cls, self)
        assert pos + 4 + 3 >= len(rawdata)
        return message

    def serialize_cdr(
        self,
        message: object,
        typename: str,
        *,
        little_endian: bool = sys.byteorder == 'little',
    ) -> memoryview:
        """Serialize message object to bytes.

        Args:
            message: Message object.
            typename: Message type name.
            little_endian: Should use little endianness.
            typestore: Type store.

        Returns:
            Serialized bytes.

        """
        msgdef = self.get_msgdef(typename)
        size = 4 + msgdef.getsize_cdr(0, message, self)
        rawdata = memoryview(bytearray(size))
        pack_into('BB', rawdata, 0, 0, little_endian)

        func = msgdef.serialize_cdr_le if little_endian else msgdef.serialize_cdr_be

        pos = func(rawdata[4:], 0, message, self)
        assert pos + 4 == size
        return rawdata.toreadonly()

    def deserialize_ros1(self, rawdata: bytes | memoryview, typename: str) -> object:
        """Deserialize raw data into a message object.

        Args:
            rawdata: Serialized data.
            typename: Message type name.
            typestore: Type store.

        Returns:
            Deserialized message object.

        """
        msgdef = self.get_msgdef(typename)
        func = msgdef.deserialize_ros1
        message, pos = func(rawdata, 0, msgdef.cls, self)
        assert pos == len(rawdata)
        return message

    def serialize_ros1(self, message: object, typename: str) -> memoryview:
        """Serialize message object to bytes.

        Args:
            message: Message object.
            typename: Message type name.
            typestore: Type store.

        Returns:
            Serialized bytes.

        """
        msgdef = self.get_msgdef(typename)
        size = msgdef.getsize_ros1(0, message, self)
        rawdata = memoryview(bytearray(size))
        func = msgdef.serialize_ros1
        pos = func(rawdata, 0, message, self)
        assert pos == size
        return rawdata.toreadonly()

    def ros1_to_cdr(self, raw: bytes | memoryview, typename: str) -> memoryview:
        """Convert serialized ROS1 message directly to CDR.

        This should be reasonably fast as conversions happen on a byte-level
        without going through deserialization and serialization.

        Args:
            raw: ROS1 serialized message.
            typename: Message type name.
            typestore: Type store.

        Returns:
            CDR serialized message.

        """
        msgdef = self.get_msgdef(typename)

        ipos, opos = msgdef.getsize_ros1_to_cdr(raw, 0, None, 0, self)
        assert ipos == len(raw)

        size = 4 + opos
        rawdata = memoryview(bytearray(size))
        pack_into('BB', rawdata, 0, 0, 1)

        ipos, opos = msgdef.ros1_to_cdr(raw, 0, rawdata[4:], 0, self)
        assert ipos == len(raw)
        assert opos + 4 == size
        return rawdata.toreadonly()

    def cdr_to_ros1(self, raw: bytes | memoryview, typename: str) -> memoryview:
        """Convert serialized CDR message directly to ROS1.

        This should be reasonably fast as conversions happen on a byte-level
        without going through deserialization and serialization.

        Args:
            raw: CDR serialized message.
            typename: Message type name.
            typestore: Type store.

        Returns:
            ROS1 serialized message.

        """
        assert raw[1] == 1, 'Message byte order is not little endian'

        msgdef = self.get_msgdef(typename)

        ipos, opos = msgdef.getsize_cdr_to_ros1(raw[4:], 0, None, 0, self)
        assert ipos + 4 + 3 >= len(raw)

        size = opos
        rawdata = memoryview(bytearray(size))

        ipos, opos = msgdef.cdr_to_ros1(raw[4:], 0, rawdata, 0, self)
        assert ipos + 4 + 3 >= len(raw)
        assert opos == size
        return rawdata.toreadonly()

    def register(self, typs: Typesdict) -> None:
        """Register types in the store.

        Args:
            typs: Dictionary mapping message typenames to parsetrees.

        Raises:
            TypesysError: Type already present with different definition.

        """
        code = generate_python_code(typs)
        name = 'rosbags.usertypes'
        spec = spec_from_loader(name, loader=None)
        assert spec
        module = module_from_spec(spec)
        sys.modules[name] = module
        exec(code, module.__dict__)  # noqa: S102
        fielddefs: Typesdict = module.FIELDDEFS

        for name, (_, fields) in fielddefs.items():
            if have := self.fielddefs.get(name):
                _, have_fields = have
                have_fields = [(x[0].lower(), x[1]) for x in have_fields]
                new_fields = [(x[0].lower(), x[1]) for x in fields]
                if have_fields != new_fields:
                    msg = f'Type {name!r} is already present with different definition.'
                    raise TypesysError(msg)

        for name in fielddefs.keys() - self.fielddefs.keys():
            self.fielddefs[name] = fielddefs[name][:2]
            self.types[name] = getattr(module, name.replace('/', '__'))

    def get_msgdef(self, typename: str) -> Msgdef[object]:
        """Retrieve message definition for typename.

        Message definitions are cached globally and generated as needed.

        Args:
            typename: Msgdef type name to load.
            typestore: Type store.

        Returns:
            Message definition.

        """
        if typename not in self.cache:
            entries = self.fielddefs[typename][1]

            getsize_cdr, size_cdr = generate_getsize_cdr(entries, self)
            getsize_ros1, size_ros1 = generate_getsize_ros1(entries, self)

            self.cache[typename] = Msgdef(
                typename,
                entries,
                self.types[typename],
                size_cdr,
                getsize_cdr,
                generate_serialize_cdr(entries, self, 'le'),
                generate_serialize_cdr(entries, self, 'be'),
                generate_deserialize_cdr(entries, self, 'le'),
                generate_deserialize_cdr(entries, self, 'be'),
                size_ros1,
                getsize_ros1,
                generate_serialize_ros1(entries, self),
                generate_deserialize_ros1(entries, self),
                cast('BitcvtSize', generate_ros1_to_cdr(entries, typename, self, copy=False)),
                cast('Bitcvt', generate_ros1_to_cdr(entries, typename, self, copy=True)),
                cast('BitcvtSize', generate_cdr_to_ros1(entries, typename, self, copy=False)),
                cast('Bitcvt', generate_cdr_to_ros1(entries, typename, self, copy=True)),
            )
        return self.cache[typename]

    def gendefhash(
        self,
        typename: str,
        subdefs: dict[str, tuple[str, str]],
        ros_version: int = 1,
    ) -> tuple[str, str]:
        """Generate message definition and hash for type.

        The subdefs argument will be filled with child definitions.

        Args:
            typename: Name of type to generate definition for.
            subdefs: Child definitions.
            ros_version: ROS version number.

        Returns:
            Message definition and hash.

        Raises:
            TypesysError: Type does not exist.

        """
        typemap = (
            {'builtin_interfaces/msg/Time': 'time', 'builtin_interfaces/msg/Duration': 'duration'}
            if ros_version == 1
            else {}
        )

        deftext: list[str] = []
        hashtext: list[str] = []
        if typename not in self.fielddefs:
            msg = f'Type {typename!r} is unknown.'
            raise TypesysError(msg)

        for name, typ, value in self.fielddefs[typename][0]:
            stripped_name = name.rstrip('_')
            deftext.append(f'{typ} {stripped_name}={value}')
            hashtext.append(f'{typ} {stripped_name}={value}')

        for name, desc in self.fielddefs[typename][1]:
            if name == 'structure_needs_at_least_one_member':
                continue
            stripped_name = name.rstrip('_')
            if desc[0] == Nodetype.BASE:
                argname: str
                argname, arglimit = desc[1]
                if argname == 'string':
                    argname = f'string<={arglimit}' if arglimit else 'string'
                deftext.append(f'{argname} {stripped_name}')
                hashtext.append(f'{argname} {stripped_name}')
            elif desc[0] == int(Nodetype.NAME):
                args = desc[1]
                assert isinstance(args, str)
                subname = args
                if subname in typemap:
                    deftext.append(f'{typemap[subname]} {stripped_name}')
                    hashtext.append(f'{typemap[subname]} {stripped_name}')
                else:
                    if subname not in subdefs:
                        subdefs[subname] = ('', '')
                        subdefs[subname] = self.gendefhash(subname, subdefs, ros_version)
                    deftext.append(f'{denormalize_msgtype(subname)} {stripped_name}')
                    hashtext.append(f'{subdefs[subname][1]} {stripped_name}')
            else:
                assert desc[0] in {Nodetype.ARRAY, Nodetype.SEQUENCE}
                assert isinstance(desc[1], tuple)
                subdesc, num = desc[1]
                isubname: tuple[str, int] | str
                isubtype, isubname = subdesc
                count = (
                    '' if num == 0 else str(num) if desc[0] == int(Nodetype.ARRAY) else f'<={num}'
                )
                if isubtype == int(Nodetype.BASE):
                    if isubname[0] == 'string':
                        isubname = (f'string<={isubname[1]}' if isubname[1] else 'string', 0)
                    deftext.append(f'{isubname[0]}[{count}] {stripped_name}')
                    hashtext.append(f'{isubname[0]}[{count}] {stripped_name}')
                elif isubname in typemap:
                    assert isinstance(isubname, str)
                    deftext.append(f'{typemap[isubname]}[{count}] {stripped_name}')
                    hashtext.append(f'{typemap[isubname]}[{count}] {stripped_name}')
                else:
                    assert isinstance(isubname, str)
                    if isubname not in subdefs:
                        subdefs[isubname] = ('', '')
                        subdefs[isubname] = self.gendefhash(isubname, subdefs, ros_version)
                    deftext.append(f'{denormalize_msgtype(isubname)}[{count}] {stripped_name}')
                    hashtext.append(f'{subdefs[isubname][1]} {stripped_name}')

        deftext.append('')
        return '\n'.join(deftext), md5('\n'.join(hashtext).encode()).hexdigest()  # noqa: S324

    def generate_msgdef(
        self,
        typename: str,
        ros_version: int = 1,
    ) -> tuple[str, str]:
        """Generate message definition for type.

        Args:
            typename: Name of type to generate definition for.
            typestore: Custom type store.
            ros_version: ROS version number.

        Returns:
            Message definition.

        """
        subdefs: dict[str, tuple[str, str]] = {}
        msgdef, md5sum = self.gendefhash(typename, subdefs, ros_version)

        msgdef = ''.join(
            [
                msgdef,
                *[f'{"=" * 80}\nMSG: {denormalize_msgtype(k)}\n{v[0]}' for k, v in subdefs.items()],
            ],
        )

        return msgdef, md5sum

    def hash_rihs01(self, typ: str) -> str:
        """Hash message definition.

        Args:
            typ: Message type name.

        Returns:
            Hash value.

        """

        def get_field(name: str, desc: FieldDesc) -> Field:
            increment = 0
            capacity = 0
            string_capacity = 0
            subtype = ''
            if desc[0] == Nodetype.ARRAY:
                increment = 48
                capacity = desc[1][1]
                typ, rest = desc[1][0]
            elif desc[0] == Nodetype.SEQUENCE:
                count = desc[1][1]
                if count:
                    increment = 96
                    capacity = count
                else:
                    increment = 144
                typ, rest = desc[1][0]
            else:
                typ, rest = desc

            if typ == Nodetype.NAME:
                tid = increment + 1
                assert isinstance(rest, str)
                subtype = rest
                _ = get_struct(subtype)
            elif rest[0] == 'string' and rest[1]:
                assert isinstance(rest[1], int)
                string_capacity = rest[1]
                tid = increment + TIDMAP['bounded_string']
            else:
                assert isinstance(rest[0], str)
                tid = increment + TIDMAP[rest[0]]

            return {
                'name': name,
                'type': {
                    'type_id': tid,
                    'capacity': capacity,
                    'string_capacity': string_capacity,
                    'nested_type_name': subtype,
                },
            }

        struct_cache: dict[str, Struct] = {}

        def get_struct(typ: str) -> Struct:
            if typ not in struct_cache:
                struct_cache[typ] = {
                    'type_name': typ,
                    'fields': list(
                        starmap(
                            get_field,
                            self.fielddefs[typ][1]
                            or [
                                (
                                    'structure_needs_at_least_one_member',
                                    (Nodetype.BASE, ('uint8', 0)),
                                )
                            ],
                        )
                    ),
                }
            return struct_cache[typ]

        dct = {
            'type_description': get_struct(typ),
            'referenced_type_descriptions': [
                y for x, y in sorted(struct_cache.items()) if x != typ
            ],
        }

        digest = sha256(json.dumps(dct).encode()).hexdigest()
        return f'RIHS01_{digest}'



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Message Registration Tests."""

import numpy as np
import pytest

from rosbags.interfaces import Nodetype
from rosbags.typesys import Stores, TypesysError, get_typestore
from rosbags.typesys.msg import get_types_from_msg
from rosbags.typesys.store import Typestore


def test_raw_typestore_is_empty() -> None:
    """Test typestore is empty."""
    store = Typestore()
    assert not store.fielddefs


def test_register_with_new_type() -> None:
    """Test typestore registers new types."""
    store = get_typestore(Stores.LATEST)

    assert 'foo' not in store.fielddefs
    store.register({'foo': ([], [('b', (Nodetype.BASE, ('bool', 0)))])})
    assert 'foo' in store.fielddefs


def test_register_with_same_as_existing_type() -> None:
    """Test typestore accepts identical type."""
    store = get_typestore(Stores.LATEST)

    typename = 'builtin_interfaces/msg/Time'

    assert typename in store.fielddefs
    store.register(
        {
            typename: (
                [],
                [
                    ('sec', (Nodetype.BASE, ('int32', 0))),
                    ('nanosec', (Nodetype.BASE, ('uint32', 0))),
                ],
            )
        }
    )


def test_register_with_different_from_existing_type() -> None:
    """Test typestore rejects different type."""
    store = get_typestore(Stores.LATEST)

    typename = 'builtin_interfaces/msg/Time'

    assert typename in store.fielddefs
    with pytest.raises(TypesysError, match='different definition'):
        store.register(
            {
                typename: (
                    [],
                    [
                        ('sec', (Nodetype.BASE, ('int32', 0))),
                        ('nsec', (Nodetype.BASE, ('uint32', 0))),
                    ],
                )
            }
        )


@pytest.mark.usefixtures('_comparable')
def test_registered_types_support_serde() -> None:
    """Test registered types allow serde operstions."""
    store = get_typestore(Stores.LATEST)

    typename_s_64_64 = 'test_msgs/msg/static_64_64'
    msgdef_s_64_64 = """
    uint64[2] u64
    """

    typename_s_64_16 = 'test_msgs/msg/static_64_16'
    msgdef_s_64_16 = """
    uint64 u64
    uint16 u16
    """

    typename_s_16_64 = 'test_msgs/msg/static_16_64'
    msgdef_s_16_64 = """
    uint16 u16
    uint64 u64
    """

    typename_d_64_64 = 'test_msgs/msg/dynamic_64_64'
    msgdef_d_64_64 = """
    uint64[] u64
    """

    typename_d_64_b_64 = 'test_msgs/msg/dynamic_64_b_64'
    msgdef_d_64_b_64 = """
    uint64 u64
    bool b
    float64 f64
    """

    typename_d_64_s = 'test_msgs/msg/dynamic_64_s'
    msgdef_d_64_s = """
    uint64 u64
    string s
    """

    typename_d_s_64 = 'test_msgs/msg/dynamic_s_64'
    msgdef_d_s_64 = """
    string s
    uint64 u64
    """

    typename_custom = 'test_msgs/msg/custom'
    msgdef_custom = """
    string base_str
    float32 base_f32
    test_msgs/msg/static_64_64 msg_s66
    test_msgs/msg/static_64_16 msg_s61
    test_msgs/msg/static_16_64 msg_s16
    test_msgs/msg/dynamic_64_64 msg_d66
    test_msgs/msg/dynamic_64_b_64 msg_d6b6
    test_msgs/msg/dynamic_64_s msg_d6s
    test_msgs/msg/dynamic_s_64 msg_ds6

    string[2] arr_base_str
    float32[2] arr_base_f32
    test_msgs/msg/static_64_64[2] arr_msg_s66
    test_msgs/msg/static_64_16[2] arr_msg_s61
    test_msgs/msg/static_16_64[2] arr_msg_s16
    test_msgs/msg/dynamic_64_64[2] arr_msg_d66
    test_msgs/msg/dynamic_64_b_64[2] arr_msg_d6b6
    test_msgs/msg/dynamic_64_s[2] arr_msg_d6s
    test_msgs/msg/dynamic_s_64[2] arr_msg_ds6

    string[] seq_base_str
    float32[] seq_base_f32
    test_msgs/msg/static_64_64[] seq_msg_s66
    test_msgs/msg/static_64_16[] seq_msg_s61
    test_msgs/msg/static_16_64[] seq_msg_s16
    test_msgs/msg/dynamic_64_64[] seq_msg_d66
    test_msgs/msg/dynamic_64_b_64[] seq_msg_d6b6
    test_msgs/msg/dynamic_64_s[] seq_msg_d6s
    test_msgs/msg/dynamic_s_64[] seq_msg_ds6
    """

    store.register(get_types_from_msg(msgdef_s_64_64, typename_s_64_64))
    store.register(get_types_from_msg(msgdef_s_64_16, typename_s_64_16))
    store.register(get_types_from_msg(msgdef_s_16_64, typename_s_16_64))
    store.register(get_types_from_msg(msgdef_d_64_64, typename_d_64_64))
    store.register(get_types_from_msg(msgdef_d_64_b_64, typename_d_64_b_64))
    store.register(get_types_from_msg(msgdef_d_64_s, typename_d_64_s))
    store.register(get_types_from_msg(msgdef_d_s_64, typename_d_s_64))
    store.register(get_types_from_msg(msgdef_custom, typename_custom))

    static_64_64 = store.types[typename_s_64_64]
    static_64_16 = store.types[typename_s_64_16]
    static_16_64 = store.types[typename_s_16_64]
    dynamic_64_64 = store.types[typename_d_64_64]
    dynamic_64_b_64 = store.types[typename_d_64_b_64]
    dynamic_64_s = store.types[typename_d_64_s]
    dynamic_s_64 = store.types[typename_d_s_64]
    custom = store.types[typename_custom]

    msg = custom(
        'str',
        1.5,
        static_64_64(np.array([64, 64], dtype=np.uint64)),
        static_64_16(64, 16),
        static_16_64(16, 64),
        dynamic_64_64(np.array([33, 33], dtype=np.uint64)),
        dynamic_64_b_64(64, b=True, f64=1.25),
        dynamic_64_s(64, 's'),
        dynamic_s_64('s', 64),
        # arrays
        ['str_1', ''],
        np.array([1.5, 0.75], dtype=np.float32),
        [
            static_64_64(np.array([64, 64], dtype=np.uint64)),
            static_64_64(np.array([64, 64], dtype=np.uint64)),
        ],
        [static_64_16(64, 16), static_64_16(64, 16)],
        [static_16_64(16, 64), static_16_64(16, 64)],
        [
            dynamic_64_64(np.array([33, 33], dtype=np.uint64)),
            dynamic_64_64(np.array([33, 33], dtype=np.uint64)),
        ],
        [dynamic_64_b_64(64, b=True, f64=1.25), dynamic_64_b_64(64, b=True, f64=1.25)],
        [dynamic_64_s(64, 's'), dynamic_64_s(64, 's')],
        [dynamic_s_64('s', 64), dynamic_s_64('s', 64)],
        # sequences
        ['str_1', ''],
        np.array([1.5, 0.75], dtype=np.float32),
        [
            static_64_64(np.array([64, 64], dtype=np.uint64)),
            static_64_64(np.array([64, 64], dtype=np.uint64)),
        ],
        [static_64_16(64, 16), static_64_16(64, 16)],
        [static_16_64(16, 64), static_16_64(16, 64)],
        [
            dynamic_64_64(np.array([33, 33], dtype=np.uint64)),
            dynamic_64_64(np.array([33, 33], dtype=np.uint64)),
        ],
        [dynamic_64_b_64(64, b=True, f64=1.25), dynamic_64_b_64(64, b=True, f64=1.25)],
        [dynamic_64_s(64, 's'), dynamic_64_s(64, 's')],
        [dynamic_s_64('s', 64), dynamic_s_64('s', 64)],
    )

    res = store.deserialize_cdr(store.serialize_cdr(msg, typename_custom), typename_custom)
    assert res == msg

    res = store.deserialize_ros1(store.serialize_ros1(msg, typename_custom), typename_custom)
    assert res == msg



"""Example: Creating a Custom Typestore.

This module defines a global variable ``nmea_typestore``, which extends
the default Jazzy typestore by adding one additional message type.

If this file is saved as ``my_typestores.py``, the fully qualified
reference to the typestore is ``my_typestores:nmea_typestore``.

"""

from __future__ import annotations

from rosbags.typesys import Stores, get_typestore
from rosbags.typesys.msg import get_types_from_msg

NMEA_SENTENCE_MSG = """
std_msgs/msg/Header header
string sentence
"""

nmea_typestore = get_typestore(Stores.ROS2_JAZZY)
nmea_typestore.register(get_types_from_msg(NMEA_SENTENCE_MSG, 'nmea_msgs/msg/Sentence'))



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Writer Tests."""

from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest

from rosbags.rosbag1 import Writer, WriterError
from rosbags.typesys import Stores, get_typestore

if TYPE_CHECKING:
    from pathlib import Path


def test_no_overwrite(tmp_path: Path) -> None:
    """Test writer does not touch existing files."""
    path = tmp_path / 'test.bag'
    _ = path.write_text('foo')
    with pytest.raises(WriterError, match='exists'):
        Writer(path).open()
    path.unlink()

    writer = Writer(path)
    _ = path.write_text('foo')
    with pytest.raises(WriterError, match='exists'):
        writer.open()


def test_empty(tmp_path: Path) -> None:
    """Test empty bag."""
    path = tmp_path / 'test.bag'

    with Writer(path):
        pass
    data = path.read_bytes()
    assert len(data) == 13 + 4096


def test_add_connection(tmp_path: Path) -> None:
    """Test adding of connections."""
    store = get_typestore(Stores.LATEST)

    path = tmp_path / 'test.bag'

    with pytest.raises(WriterError, match='not opened'):
        _ = Writer(path).add_connection(
            '/foo',
            'test_msgs/msg/Test',
            msgdef='MESSAGE_DEFINITION',
            md5sum='HASH',
        )

    with Writer(path) as writer:
        res = writer.add_connection(
            '/foo',
            'test_msgs/msg/Test',
            msgdef='MESSAGE_DEFINITION',
            md5sum='HASH',
        )
        assert res.id == 0
    data = path.read_bytes()
    assert data.count(b'MESSAGE_DEFINITION') == 2
    assert data.count(b'HASH') == 2
    path.unlink()

    with Writer(path) as writer:
        res = writer.add_connection('/foo', 'std_msgs/msg/Int8', typestore=store)
        assert res.id == 0
    data = path.read_bytes()
    assert data.count(b'int8 data') == 2
    assert data.count(b'27ffa0c9c4b8fb8492252bcad9e5c57b') == 2
    path.unlink()

    with Writer(path) as writer:
        _ = writer.add_connection(
            '/foo',
            'test_msgs/msg/Test',
            msgdef='MESSAGE_DEFINITION',
            md5sum='HASH',
        )
        with pytest.raises(WriterError, match='can only be added once'):
            _ = writer.add_connection(
                '/foo',
                'test_msgs/msg/Test',
                msgdef='MESSAGE_DEFINITION',
                md5sum='HASH',
            )
    path.unlink()

    with Writer(path) as writer:
        res1 = writer.add_connection(
            '/foo',
            'test_msgs/msg/Test',
            msgdef='MESSAGE_DEFINITION',
            md5sum='HASH',
        )
        res2 = writer.add_connection(
            '/foo',
            'test_msgs/msg/Test',
            msgdef='MESSAGE_DEFINITION',
            md5sum='HASH',
            callerid='src',
        )
        res3 = writer.add_connection(
            '/foo',
            'test_msgs/msg/Test',
            msgdef='MESSAGE_DEFINITION',
            md5sum='HASH',
            latching=1,
        )
        assert (res1.id, res2.id, res3.id) == (0, 1, 2)


def test_write_errors(tmp_path: Path) -> None:
    """Test write errors."""
    path = tmp_path / 'test.bag'

    with pytest.raises(WriterError, match='not opened'):
        Writer(path).write(Mock(), 42, b'DEADBEEF')

    with Writer(path) as writer, pytest.raises(WriterError, match='is no connection'):
        writer.write(Mock(), 42, b'DEADBEEF')
    path.unlink()


def test_write_simple(tmp_path: Path) -> None:
    """Test writing of messages."""
    path = tmp_path / 'test.bag'

    with Writer(path) as writer:
        conn_foo = writer.add_connection(
            '/foo',
            'test_msgs/msg/Test',
            msgdef='MESSAGE_DEFINITION',
            md5sum='HASH',
        )
        conn_latching = writer.add_connection(
            '/foo',
            'test_msgs/msg/Test',
            msgdef='MESSAGE_DEFINITION',
            md5sum='HASH',
            latching=1,
        )
        conn_bar = writer.add_connection(
            '/bar',
            'test_msgs/msg/Bar',
            msgdef='OTHER_DEFINITION',
            md5sum='HASH',
            callerid='src',
        )
        _ = writer.add_connection(
            '/baz',
            'test_msgs/msg/Baz',
            msgdef='NEVER_WRITTEN',
            md5sum='HASH',
        )

        writer.write(conn_foo, 42, b'DEADBEEF')
        writer.write(conn_latching, 42, b'DEADBEEF')
        writer.write(conn_bar, 43, b'SECRET')
        writer.write(conn_bar, 43, b'SUBSEQUENT')

    res = path.read_bytes()
    assert res.count(b'op=\x05') == 1
    assert res.count(b'op=\x06') == 1
    assert res.count(b'MESSAGE_DEFINITION') == 4
    assert res.count(b'latching=1') == 2
    assert res.count(b'OTHER_DEFINITION') == 2
    assert res.count(b'callerid=src') == 2
    assert res.count(b'NEVER_WRITTEN') == 2
    assert res.count(b'DEADBEEF') == 2
    assert res.count(b'SECRET') == 1
    assert res.count(b'SUBSEQUENT') == 1
    path.unlink()

    with Writer(path) as writer:
        writer.chunk_threshold = 256
        conn_foo = writer.add_connection(
            '/foo',
            'test_msgs/msg/Test',
            msgdef='MESSAGE_DEFINITION',
            md5sum='HASH',
        )
        conn_latching = writer.add_connection(
            '/foo',
            'test_msgs/msg/Test',
            msgdef='MESSAGE_DEFINITION',
            md5sum='HASH',
            latching=1,
        )
        conn_bar = writer.add_connection(
            '/bar',
            'test_msgs/msg/Bar',
            msgdef='OTHER_DEFINITION',
            md5sum='HASH',
            callerid='src',
        )
        _ = writer.add_connection(
            '/baz', 'test_msgs/msg/Baz', msgdef='NEVER_WRITTEN', md5sum='HASH'
        )

        writer.write(conn_foo, 42, b'DEADBEEF')
        writer.write(conn_latching, 42, b'DEADBEEF')
        writer.write(conn_bar, 43, b'SECRET')
        writer.write(conn_bar, 43, b'SUBSEQUENT')

    res = path.read_bytes()
    assert res.count(b'op=\x05') == 2
    assert res.count(b'op=\x06') == 2
    assert res.count(b'MESSAGE_DEFINITION') == 4
    assert res.count(b'latching=1') == 2
    assert res.count(b'OTHER_DEFINITION') == 2
    assert res.count(b'callerid=src') == 2
    assert res.count(b'NEVER_WRITTEN') == 2
    assert res.count(b'DEADBEEF') == 2
    assert res.count(b'SECRET') == 1
    assert res.count(b'SUBSEQUENT') == 1
    path.unlink()


def test_compression_errors(tmp_path: Path) -> None:
    """Test compression modes."""
    path = tmp_path / 'test.bag'
    with Writer(path) as writer, pytest.raises(WriterError, match='already open'):
        writer.set_compression(writer.CompressionFormat.BZ2)


@pytest.mark.parametrize('fmt', [None, Writer.CompressionFormat.BZ2, Writer.CompressionFormat.LZ4])
def test_compression_modes(tmp_path: Path, fmt: Writer.CompressionFormat | None) -> None:
    """Test compression modes."""
    store = get_typestore(Stores.LATEST)
    path = tmp_path / 'test.bag'
    writer = Writer(path)
    if fmt:
        writer.set_compression(fmt)
    with writer:
        conn = writer.add_connection('/foo', 'std_msgs/msg/Int8', typestore=store)
        writer.write(conn, 42, b'\x42')
    data = path.read_bytes()
    assert data.count(f'compression={fmt.name.lower() if fmt else "none"}'.encode()) == 1


@pytest.mark.parametrize('fmt', [None, Writer.CompressionFormat.BZ2, Writer.CompressionFormat.LZ4])
def test_chunksize_is_correct(tmp_path: Path, fmt: Writer.CompressionFormat | None) -> None:
    """Test chunksize is correct."""
    store = get_typestore(Stores.LATEST)
    path = tmp_path / 'test1.bag'
    writer = Writer(path)
    if fmt:
        writer.set_compression(fmt)
    with writer:
        conn = writer.add_connection('/foo', 'std_msgs/msg/Int8', typestore=store)
        writer.write(conn, 42, b'\x42')
    data = path.read_bytes()
    assert b'size=\xca\x00\x00\x00' in data

    path = tmp_path / 'test2.bag'
    writer = Writer(path)
    if fmt:
        writer.set_compression(fmt)
    with writer:
        conn = writer.add_connection('/foo', 'std_msgs/msg/Int8', typestore=store)
        writer.write(conn, 42, b'\x42')
        writer.write(conn, 43, b'\x43')
    data = path.read_bytes()
    assert b'size=\xf9\x00\x00\x00' in data


def test_deprecations(tmp_path: Path) -> None:
    """Test writer deprecations."""
    bag = Writer(tmp_path / 'bag')
    with bag, pytest.deprecated_call():
        _ = bag.add_connection('/foo', 'std_msgs/msg/Empty')



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""ROS1 Generator Tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

from rosbags.interfaces import Nodetype
from rosbags.serde.ros1 import (
    generate_cdr_to_ros1,
    generate_deserialize_ros1,
    generate_getsize_ros1,
    generate_ros1_to_cdr,
    generate_serialize_ros1,
)
from rosbags.typesys import Stores, get_typestore

if TYPE_CHECKING:
    from rosbags.interfaces.typing import Fielddefs

ALL_COMBINATIONS: Fielddefs = [
    ('structure_needs_at_least_one_member', (Nodetype.BASE, ('uint8', 0))),
    ('fmsg', (Nodetype.NAME, 'f_size')),
    ('vmsg', (Nodetype.NAME, 'v_size')),
    ('u64', (Nodetype.BASE, ('uint64', 0))),
    ('str', (Nodetype.BASE, ('string', 0))),
    ('a_fmsg', (Nodetype.ARRAY, ((Nodetype.NAME, 'f_size'), 2))),
    ('a_vmsg', (Nodetype.ARRAY, ((Nodetype.NAME, 'v_size'), 2))),
    ('a_u64', (Nodetype.ARRAY, ((Nodetype.BASE, ('uint64', 0)), 2))),
    ('a_str', (Nodetype.ARRAY, ((Nodetype.BASE, ('string', 0)), 2))),
    ('s_fmsg', (Nodetype.SEQUENCE, ((Nodetype.NAME, 'f_size'), 2))),
    ('s_vmsg', (Nodetype.SEQUENCE, ((Nodetype.NAME, 'v_size'), 2))),
    ('s_u64', (Nodetype.SEQUENCE, ((Nodetype.BASE, ('uint64', 0)), 2))),
    ('s_str', (Nodetype.SEQUENCE, ((Nodetype.BASE, ('string', 0)), 2))),
]


def test_ros1_generator() -> None:
    """Test ros1 code generators do not throw."""
    store = get_typestore(Stores.EMPTY)
    store.register(
        {
            'f_size': ([], [('u64', (Nodetype.BASE, ('uint64', 0)))]),
            'v_size': ([], [('s_u8', (Nodetype.SEQUENCE, ((Nodetype.BASE, ('uint8', 0)), 0)))]),
        }
    )
    _ = generate_getsize_ros1(ALL_COMBINATIONS, store)
    _ = generate_deserialize_ros1(ALL_COMBINATIONS, store)
    _ = generate_serialize_ros1(ALL_COMBINATIONS, store)

    for typename, copy in zip(('foo', 'std_msgs/msg/Header'), (True, False), strict=True):
        _ = generate_cdr_to_ros1(ALL_COMBINATIONS, typename, store, copy=copy)
        _ = generate_ros1_to_cdr(ALL_COMBINATIONS, typename, store, copy=copy)



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""MSG Message Definition Parser Tests."""

import pytest

from rosbags.interfaces import Nodetype
from rosbags.typesys import Stores, TypesysError, get_types_from_msg, get_typestore

MSG = """
# comment

bool b=true
int32 global=42
float32 f=1.33
string str= foo bar\t

std_msgs/Header header
std_msgs/msg/Bool bool
test_msgs/Bar sibling
float64 base
float64[] seq1
float64[] seq2
float64[4] array
"""

MSG_BOUNDS = """
int32[] unbounded_integer_array
int32[5] five_integers_array
int32[<=5] up_to_five_integers_array

string string_of_unbounded_size
string<=10 up_to_ten_characters_string

string[<=5] up_to_five_unbounded_strings
string<=10[] unbounded_array_of_string_up_to_ten_characters_each
string<=10[<=5] up_to_five_strings_up_to_ten_characters_each
"""

MSG_DEFAULTS = """
bool b false
uint8 i 42
uint8 o 0377
uint8 h 0xff
float32 y -314.15e-2
string name1 "John"
string name2 'Ringo'
int32[] samples [-200, -100, 0, 100, 200]
"""

MULTI_MSG = """
std_msgs/Header header
byte b
char c
Other[] o

================================================================================
MSG: std_msgs/Header
time time

================================================================================
MSG: test_msgs/Other
uint64[3] Header
uint32 static = 42
"""

CSTRING_CONFUSION_MSG = """
std_msgs/Header header
string s

================================================================================
MSG: std_msgs/Header
time time
"""

RELSIBLING_MSG = """
Header header
Other other
"""

KEYWORD_MSG = """
bool return=true
uint64 yield
"""

NAME_COLLISION_MSG = """
std_msgs/Int s_int
Int c_int

================================================================================
MSG: collision_msgs/Int
int16 value
================================================================================
MSG: std_msgs/Int
int8 data
"""

WITH_LITERALS_MSG = """
float32 both 0.0
float32 integer_only 0.
float32 fraction_only .0
float32 scientific_pos 1e1
float32 scientific_neg 1e-1
"""

ACTION_DEFINITION = """
unique_identifier_msgs/msg/UUID goal_id
control_msgs/action/FollowJointTrajectory_Feedback feedback
"""


def test_msg_parser_raises_on_bad_definition() -> None:
    """Test msg parser raises on bad definition."""
    with pytest.raises(TypesysError, match='Could not parse'):
        _ = get_types_from_msg('invalid', 'test_msgs/msg/Foo')


def test_msg_parser_accepts_empty_definition() -> None:
    """Test msg parser accepts empty message definition."""
    ret = get_types_from_msg('', 'std_msgs/msg/Empty')
    assert ret == {'std_msgs/msg/Empty': ([], [])}


def test_msg_parser_accepts_single_msg() -> None:
    """Test msg parser accepts single message."""
    ret = get_types_from_msg(MSG, 'test_msgs/msg/Foo')
    get_typestore(Stores.EMPTY).register(ret)

    assert 'test_msgs/msg/Foo' in ret
    consts, fields = ret['test_msgs/msg/Foo']
    assert consts == [
        ('b', 'bool', True),
        ('global_', 'int32', 42),
        ('f', 'float32', 1.33),
        ('str', 'string', 'foo bar'),
    ]
    assert fields[0][0] == 'header'
    assert fields[0][1][1] == 'std_msgs/msg/Header'
    assert fields[1][0] == 'bool'
    assert fields[1][1][1] == 'std_msgs/msg/Bool'
    assert fields[2][0] == 'sibling'
    assert fields[2][1][1] == 'test_msgs/msg/Bar'
    assert fields[3][1][0] == int(Nodetype.BASE)
    assert fields[4][1][0] == int(Nodetype.SEQUENCE)
    assert fields[5][1][0] == int(Nodetype.SEQUENCE)
    assert fields[6][1][0] == int(Nodetype.ARRAY)


def test_msg_parser_processes_bounds() -> None:
    """Test msg parser processes string and sequence bounds."""
    ret = get_types_from_msg(MSG_BOUNDS, 'test_msgs/msg/Foo')
    get_typestore(Stores.EMPTY).register(ret)

    assert ret == {
        'test_msgs/msg/Foo': (
            [],
            [
                ('unbounded_integer_array', (4, ((1, ('int32', 0)), 0))),
                ('five_integers_array', (3, ((1, ('int32', 0)), 5))),
                ('up_to_five_integers_array', (4, ((1, ('int32', 0)), 5))),
                ('string_of_unbounded_size', (1, ('string', 0))),
                ('up_to_ten_characters_string', (1, ('string', 10))),
                ('up_to_five_unbounded_strings', (4, ((1, ('string', 0)), 5))),
                (
                    'unbounded_array_of_string_up_to_ten_characters_each',
                    (4, ((1, ('string', 10)), 0)),
                ),
                ('up_to_five_strings_up_to_ten_characters_each', (4, ((1, ('string', 10)), 5))),
            ],
        ),
    }


def test_msg_parser_accepts_field_defaults() -> None:
    """Test msg parser accepts field defaults."""
    ret = get_types_from_msg(MSG_DEFAULTS, 'test_msgs/msg/Foo')
    get_typestore(Stores.EMPTY).register(ret)

    assert ret == {
        'test_msgs/msg/Foo': (
            [],
            [
                ('b', (1, ('bool', 0))),
                ('i', (1, ('uint8', 0))),
                ('o', (1, ('uint8', 0))),
                ('h', (1, ('uint8', 0))),
                ('y', (1, ('float32', 0))),
                ('name1', (1, ('string', 0))),
                ('name2', (1, ('string', 0))),
                ('samples', (4, ((1, ('int32', 0)), 0))),
            ],
        ),
    }


def test_msg_parser_accepts_multiple_definitions() -> None:
    """Test msg parser accepts concatenated definitions."""
    ret = get_types_from_msg(MULTI_MSG, 'test_msgs/msg/Foo')
    get_typestore(Stores.EMPTY).register(ret)

    assert len(ret) == 3
    assert 'test_msgs/msg/Foo' in ret
    assert 'std_msgs/msg/Header' in ret
    assert 'test_msgs/msg/Other' in ret
    fields = ret['test_msgs/msg/Foo'][1]
    assert fields[0][1][1] == 'std_msgs/msg/Header'
    assert fields[1][1][1] == ('byte', 0)
    assert fields[2][1][1] == ('char', 0)
    consts = ret['test_msgs/msg/Other'][0]
    assert consts == [('static', 'uint32', 42)]


def test_msg_parser_does_not_confuse_string_constants() -> None:
    """Test msg parser does not confuse message separator with const string."""
    ret = get_types_from_msg(CSTRING_CONFUSION_MSG, 'test_msgs/msg/Foo')
    get_typestore(Stores.EMPTY).register(ret)

    assert len(ret) == 2
    assert 'test_msgs/msg/Foo' in ret
    assert 'std_msgs/msg/Header' in ret
    consts, fields = ret['test_msgs/msg/Foo']
    assert consts == []
    assert fields[0][1][1] == 'std_msgs/msg/Header'
    assert fields[1][1][1] == ('string', 0)


def test_msg_parser_resolves_relative_siblings() -> None:
    """Test msg parser resolves relative siblings."""
    ret = get_types_from_msg(RELSIBLING_MSG, 'test_msgs/msg/Foo')
    get_typestore(Stores.EMPTY).register(ret)

    assert ret['test_msgs/msg/Foo'][1][0][1][1] == 'std_msgs/msg/Header'
    assert ret['test_msgs/msg/Foo'][1][1][1][1] == 'test_msgs/msg/Other'

    ret = get_types_from_msg(RELSIBLING_MSG, 'rel_msgs/msg/Foo')
    get_typestore(Stores.EMPTY).register(ret)

    assert ret['rel_msgs/msg/Foo'][1][0][1][1] == 'std_msgs/msg/Header'
    assert ret['rel_msgs/msg/Foo'][1][1][1][1] == 'rel_msgs/msg/Other'


def test_msg_parser_avoids_python_keyword_collisions() -> None:
    """Test msg parser renames message fields to avoid python keyword collisions."""
    ret = get_types_from_msg(KEYWORD_MSG, 'keyword_msgs/msg/Foo')
    get_typestore(Stores.EMPTY).register(ret)

    assert ret['keyword_msgs/msg/Foo'][0][0][0] == 'return_'
    assert ret['keyword_msgs/msg/Foo'][1][0][0] == 'yield_'


def test_msg_parser_handles_name_collisions() -> None:
    """Test msg parser handles identical type names in different namespaces."""
    ret = get_types_from_msg(NAME_COLLISION_MSG, 'collision_msgs/msg/Foo')
    get_typestore(Stores.EMPTY).register(ret)

    assert ret['collision_msgs/msg/Foo'][1][0][1][1] == 'std_msgs/msg/Int'
    assert ret['collision_msgs/msg/Foo'][1][1][1][1] == 'collision_msgs/msg/Int'


def test_literal_notations() -> None:
    """Test msg parser handles different literal notations."""
    ret = get_types_from_msg(WITH_LITERALS_MSG, 'literals_msgs/msg/Foo')
    get_typestore(Stores.EMPTY).register(ret)

    assert len(ret['literals_msgs/msg/Foo'][1]) == 5


def test_actions_definitions() -> None:
    """Test msg parser action definitions."""
    ret = get_types_from_msg(
        ACTION_DEFINITION, 'control_msgs/action/FollowJointTrajectory_FeedbackMessage'
    )
    assert ret == {
        'control_msgs/action/FollowJointTrajectory_FeedbackMessage': (
            [],
            [
                ('goal_id', (Nodetype.NAME, 'unique_identifier_msgs/msg/UUID')),
                (
                    'feedback',
                    (
                        Nodetype.NAME,
                        'control_msgs/action/FollowJointTrajectory_Feedback',
                    ),
                ),
            ],
        )
    }



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Code Generator Tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

from rosbags.interfaces import Nodetype
from rosbags.typesys.codegen import generate_python_code

if TYPE_CHECKING:
    from rosbags.interfaces.typing import Typesdict


def test_generate_python_code() -> None:
    """Test python code generator."""
    defs: Typesdict = {'foo': ([], [('b', (Nodetype.BASE, ('bool', 0)))])}

    res = generate_python_code(defs)
    assert 'class foo:\n' in res

    res = generate_python_code(defs, 'empty')
    assert 'from . import empty' in res
    assert 'class foo:\n' not in res

    res = generate_python_code(defs, 'empty', add=('foo',))
    assert 'class foo:\n' in res

    res = generate_python_code(defs, 'empty', change=('foo',))
    assert 'class foo:\n' in res

    res = generate_python_code(defs, 'empty', keep=('bar',))
    assert 'bar = base.bar\n' in res



# Copyright 2016 - 2023 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Generated declinate CLI tests."""

# DO NOT EDIT THIS FILE MANUALLY

import inspect

from declinate.check import check_package
from declinate.cli import main


def test_cli_is_up_to_date() -> None:
    """Test if compiled cli is up to date."""
    res = check_package('rosbags.convert')
    assert not res, res


def test_cli() -> None:
    """Cli."""
    assert inspect.isfunction(main)



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Writer/Reader Roundtrip Tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from rosbags.rosbag1 import Reader, Writer
from rosbags.typesys import Stores, get_typestore
from rosbags.typesys.stores.ros1_noetic import std_msgs__msg__Float64 as Float64

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize('fmt', [None, *Writer.CompressionFormat])
def test_roundtrip(tmp_path: Path, fmt: Writer.CompressionFormat | None) -> None:
    """Test messages stay the same between write and read."""
    store = get_typestore(Stores.ROS1_NOETIC)

    float64 = Float64(1.25)

    path = tmp_path / 'test.bag'
    wbag = Writer(path)
    if fmt:
        wbag.set_compression(fmt)
    with wbag:
        wconnection = wbag.add_connection('/test', float64.__msgtype__, typestore=store)
        wbag.write(wconnection, 42, store.serialize_ros1(float64, float64.__msgtype__))

    rbag = Reader(path)
    with rbag:
        gen = rbag.messages()
        rconnection, _, raw = next(gen)
        assert rconnection.topic == wconnection.topic
        # Should implement: assert rconnection.msgtype == wconnection.msgtype
        assert rconnection.ext == wconnection.ext
        msg = store.deserialize_ros1(raw, rconnection.msgtype)
        assert msg == float64
        with pytest.raises(StopIteration):
            _ = next(gen)



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""CLI tool for rosbag conversion."""

from __future__ import annotations

from .cli import main

if __name__ == '__main__':
    main()



"""Example: Edit timestamps."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from rosbags.rosbag2 import Reader, Writer
from rosbags.typesys import Stores, get_typestore

if TYPE_CHECKING:
    from pathlib import Path

    from rosbags.interfaces import ConnectionExtRosbag2


def offset_timestamps(src: Path, dst: Path, offset: int) -> None:
    """Offset timestamps.

    Args:
        src: Source path.
        dst: Destination path.
        offset: Amount of nanoseconds to offset timestamps.

    """
    typestore = get_typestore(Stores.ROS2_FOXY)
    with Reader(src) as reader, Writer(dst, version=9) as writer:
        conn_map = {}
        for conn in reader.connections:
            ext = cast('ConnectionExtRosbag2', conn.ext)
            conn_map[conn.id] = writer.add_connection(
                conn.topic,
                conn.msgtype,
                typestore=typestore,
                serialization_format=ext.serialization_format,
                offered_qos_profiles=ext.offered_qos_profiles,
            )

        for conn, timestamp, data in reader.messages():
            # Adjust header timestamps, too
            msg = typestore.deserialize_cdr(data, conn.msgtype)
            if head := getattr(msg, 'header', None):
                headstamp = head.stamp.sec * 10**9 + head.stamp.nanosec + offset
                head.stamp.sec = headstamp // 10**9
                head.stamp.nanosec = headstamp % 10**9
                outdata: memoryview | bytes = typestore.serialize_cdr(msg, conn.msgtype)
            else:
                outdata = data

            writer.write(conn_map[conn.id], timestamp + offset, outdata)
